from ardent import imageData
import numpy as np

class Test_ImageMetaData:
    """Test class for ImageMetaData."""

    def test_validate_nxyz(self):

        # Test valid runs.

        valid_inputs = [
            # Scalar nxyz.
            {'nxyz':0, 'image':None}, 
            {'nxyz':1, 'image':None}, 
            {'nxyz':9, 'image':None}, 
            # nxyz from 1-D iterables.
            {'nxyz':[1, 2, 3], 'image':None}, 
            {'nxyz':np.array([1, 2, 3]), 'image':None}, 
            # Varying-dimensional image of types: scalar, list, np.ndarray.
            {'nxyz':None, 'image':5}, 
            {'nxyz':None, 'image':[1, 2]}, 
            {'nxyz':None, 'image':[[1, 2], [3, 4], [5, 6]]}, 
            {'nxyz':None, 'image':np.arange(3*4*5).reshape(3,4,5)}, 
            # nxyz and image agreement.
            {'nxyz':1, 'image':5}, 
            {'nxyz':[2], 'image':[1, 2]}, 
            {'nxyz':[3,2], 'image':[[1, 2], [3, 4], [5, 6]]}, 
            {'nxyz':np.array([3,4,5]), 'image':np.arange(3*4*5).reshape(3,4,5)}
        ]
        valid_outputs = [
            # Scalar nxyz.
            np.array([0]), 
            np.array([1]), 
            np.array([9]), 
            # nxyz from 1-D iterables.
            np.array([1, 2, 3]), 
            np.array([1, 2, 3]), 
            # Varying-dimensional image.
            np.array([1]), 
            np.array([2]), 
            np.array([3, 2]), 
            np.array([3, 4, 5]), 
            # nxyz and image agreement.
            np.array([1]), 
            np.array([2]), 
            np.array([3, 2]), 
            np.array([3, 4, 5])
        ]

        for inputs, output in zip(valid_inputs, valid_outputs):
            assert np.all(imageData.ImageMetaData._validate_nxyz(**inputs) == output)
        
        # Test invalid runs.




preset_parameters = {}

# Define registration parameter presets.
preset_parameters.update({'identity'          : dict(translational_stepsize=0,    linear_stepsize=0,     deformative_stepsize=0,    sigma_regularization=0, num_affine_only_iterations=0, num_iterations=1)})
preset_parameters.update({'clarity, mouse'    : dict(translational_stepsize=1e-5, linear_stepsize=5e-8,  deformative_stepsize=5e-1, sigma_regularization=2e1)})
preset_parameters.update({'nissl, mouse'      : dict(translational_stepsize=2e-9, linear_stepsize=1e-13, deformative_stepsize=5e-4, sigma_regularization=1e0)})
preset_parameters.update({'mri, human'        : dict(translational_stepsize=1e-9, linear_stepsize=5e-13, deformative_stepsize=5e-4, sigma_regularization=1e0)})
# preset_parameters.update({'clarity' : dict(sigma_regularization=1e1, deformative_stepsize=5e-1, linear_stepsize=2e-8, translational_stepsize=2e-5)}) # TODO: remove deprecated 'clarity' preset.


def get_registration_presets():
    """
    Get the names of all registration presets.
    
    Returns:
        dict_keys: The keys of the dictionary mapping all registration preset names to the corresponding registration parameters.
    """

    return preset_parameters.keys()


def get_registration_preset(preset:str) -> dict:
    """
    If <preset> is recognized, returns a dictionary containing the registration parameters corresponding to <preset>.
    
    Args:
        preset (str): The name of a preset keyed to a particular dictionary of registration parameters.
    
    Raises:
        NotImplementedError: Raised if <preset> is not a recognized preset name.
    
    Returns:
        dict: The registration kwargs specified by <preset>.
    """

    preset = preset.strip().lower()

    if preset in preset_parameters:
        return dict(preset_parameters[preset]) # Recast as dict to provide a freely mutable copy.
    else:
        raise NotImplementedError(f"There is no preset for '{preset}'.\n"
            f"Recognized presets include:\n{list(preset_parameters.keys())}.")



"""
Defines:
    Functions:
        _validate_scalar_to_multi(value, size=3, dtype=float)
        _validate_ndarray(array, minimum_ndim=0, required_ndim=None, dtype=None, 
            forbid_object_dtype=True, broadcast_to_shape=None)
        _validate_xyz_resolution(ndim, xyz_resolution)
        _compute_axes(shape, xyz_resolution=1, origin='center')
        _compute_coords(shape, xyz_resolution=1, origin='center')
        _multiply_by_affine(array, affine)
"""

import numpy as np

def _validate_scalar_to_multi(value, size=3, dtype=float):
    """
    If value's length is 1, upcast it to match size. 
    Otherwise, if it does not match size, raise error.

    Return a numpy array.
    """

    # Cast size to int.
    try:
        size = int(size)
    except (TypeError, ValueError):
        raise TypeError(f"size must be interpretable as an integer.\n"
            f"type(size): {type(size)}.")
    
    if size < 0:
        raise ValueError(f"size must be non-negative.\n"
            f"size: {size}.")
    
    # Cast value to np.ndarray.
    try:
        value = np.array(value, dtype)
    except ValueError:
        raise ValueError(f"value and dtype are incompatible with one another.")

    # Validate value's dimensionality and length.
    if value.ndim == 0:
        value = np.array([value])
    if value.ndim == 1:
        if len(value) == 1:
            # Upcast scalar to match size.
            value = np.full(size, value, dtype=dtype)
        elif len(value) != size:
            # value's length is incompatible with size.
            raise ValueError(f"The length of value must either be 1 or it must match size.\n"
                f"len(value): {len(value)}, size: {size}.")
    else:
        # value.ndim > 1.
        raise ValueError(f"value must not have more than 1 dimension.\n"
            f"value.ndim: {value.ndim}.")
    
    # TODO: verify that this is necessary and rewrite/remove accordingly.
    # Check for np.nan values.
    if np.any(np.isnan(value)):
        raise NotImplementedError("np.nan values encountered. What input led to this result?\n"
            "Write in an exception as appropriate.")
        raise ValueError(f"value contains inappropriate values for the chosen dtype "
            f"and thus contains np.nan values.")
            
    return value


def _validate_ndarray(array, minimum_ndim=0, required_ndim=None, dtype=None, 
forbid_object_dtype=True, broadcast_to_shape=None, reshape_to_shape=None):
    """Cast (a copy of) array to a np.ndarray if possible and return it 
    unless it is noncompliant with minimum_ndim, required_ndim, and dtype.
    
    Note:
    
    If required_ndim is None, _validate_ndarray will accept any object.
    If it is possible to cast to dtype, otherwise an exception is raised.

    If np.array(array).ndim == 0 and required_ndim == 1, array will be upcast to ndim 1.
    
    If forbid_object_dtype == True and the dtype is object, an exception is raised 
    unless object is the dtype.
    
    If a shape is provided to broadcast_to_shape, unless noncompliance is found with 
    required_ndim, array is broadcasted to that shape.
    
    if a shape is provided to reshape_to_shape, array is reshaped to that shape."""

    # Verify arguments.

    # Verify minimum_ndim.
    if not isinstance(minimum_ndim, int):
        raise TypeError(f"minimum_ndim must be of type int.\n"
            f"type(minimum_ndim): {type(minimum_ndim)}.")
    if minimum_ndim < 0:
        raise ValueError(f"minimum_ndim must be non-negative.\n"
            f"minimum_ndim: {minimum_ndim}.")

    # Verify required_ndim.
    if required_ndim is not None:
        if not isinstance(required_ndim, int):
            raise TypeError(f"required_ndim must be either None or of type int.\n"
                f"type(required_ndim): {type(required_ndim)}.")
        if required_ndim < 0:
            raise ValueError(f"required_ndim must be non-negative.\n"
                f"required_ndim: {required_ndim}.")

    # Verify dtype.
    if dtype is not None:
        if not isinstance(dtype, type):
            raise TypeError(f"dtype must be either None or a valid type.\n"
                f"type(dtype): {type(dtype)}.")

    # Validate array.

    # Cast array to np.ndarray.
    # Validate compliance with dtype.
    try:
        array = np.array(array, dtype) # Side effect: breaks alias.
    except TypeError:
        raise TypeError(f"array is of a type that is incompatible with dtype.\n"
            f"type(array): {type(array)}, dtype: {dtype}.")
    except ValueError:
        raise ValueError(f"array has a value that is incompatible with dtype.\n"
            f"array: {array}, \ntype(array): {type(array)}, dtype: {dtype}.")

    # Verify compliance with forbid_object_dtype.
    if forbid_object_dtype:
        if array.dtype == object and dtype != object:
            raise TypeError(f"Casting array to a np.ndarray produces an array of dtype object \n"
                f"while forbid_object_dtype == True and dtype != object.")

    # Validate compliance with required_ndim.
    if required_ndim is not None and array.ndim != required_ndim:
        # Upcast from ndim 0 to ndim 1 if appropriate.
        if array.ndim == 0 and required_ndim == 1:
            array = np.array([array])
        else:
            raise ValueError(f"If required_ndim is not None, array.ndim must equal it unless array.ndim == 0 and required_ndin == 1.\n"
                f"array.ndim: {array.ndim}, required_ndim: {required_ndim}.")

    # Verify compliance with minimum_ndim.
    if array.ndim < minimum_ndim:
        raise ValueError(f"array.ndim must be at least equal to minimum_ndim.\n"
            f"array.ndim: {array.ndim}, minimum_ndim: {minimum_ndim}.")
    
    # Broadcast array if appropriate.
    if broadcast_to_shape is not None:
        array = np.copy(np.broadcast_to(array=array, shape=broadcast_to_shape))

    # Reshape array if appropriate.
    if reshape_to_shape is not None:
        array = np.copy(array.reshape(reshape_to_shape))

    return array

# TODO: reverse order of arguments and propagate change throughout ardent.
def _validate_xyz_resolution(ndim, xyz_resolution):
    """Validate xyz_resolution to assure its length matches the dimensionality of image."""

    xyz_resolution = _validate_scalar_to_multi(xyz_resolution, size=ndim)

    if np.any(xyz_resolution <= 0):
        raise ValueError(f"All elements of xyz_resolution must be positive.\n"
            f"np.min(xyz_resolution): {np.min(xyz_resolution)}.")

    return xyz_resolution


def _compute_axes(shape, xyz_resolution=1, origin='center'):
    """Returns the real_axes defining an image with the given shape 
    at the given resolution as a list of numpy arrays.
    """

    # Validate shape.
    shape = _validate_ndarray(shape, dtype=int, required_ndim=1)

    # Validate xyz_resolution.
    xyz_resolution = _validate_xyz_resolution(len(shape), xyz_resolution)

    # Create axes.

    # axes is a list of arrays matching each shape element from shape, spaced by the corresponding xyz_resolution.
    axes = [np.arange(dim_size) * dim_res for dim_size, dim_res in zip(shape, xyz_resolution)]

    # List all presently recognized origin values.
    origins = ['center', 'zero']

    if origin == 'center':
        # Center each axes array to its mean.
        for xyz_index, axis in enumerate(axes):
            axes[xyz_index] -= np.mean(axis)
    elif origin == 'zero':
        # Allow each axis to increase from 0 along each dimension.
        pass
    else:
        raise NotImplementedError(f"origin must be one of these supported values: {origins}.\n"
            f"origin: {origin}.")
    
    return axes


def _compute_coords(shape, xyz_resolution=1, origin='center'):
    """Returns the real_coordinates of an image with the given shape 
    at the given resolution as a single numpy array of shape (*shape, len(shape))."""

    axes = _compute_axes(shape, xyz_resolution, origin)

    meshes = np.meshgrid(*axes, indexing='ij')

    return np.stack(meshes, axis=-1)


def _multiply_by_affine(array, affine, spatial_dimensions=3):

    arrays = []
    for dim in range(spatial_dimensions):
        arrays.append(np.sum(affine[dim, :-1] * array + affine[dim, -1], axis=-1))

    return np.stack(arrays=arrays, axis=-1)

    # Expanded for 3D:
    # return np.stack(
    #     arrays=[
    #         np.sum(affine[0, :3] * array + affine[0, -1], axis=-1), 
    #         np.sum(affine[1, :3] * array + affine[1, -1], axis=-1), 
    #         np.sum(affine[2, :3] * array + affine[2, -1], axis=-1), 
    #     ],
    #     axis=-1,
    # )



import numpy as np
from skimage.transform import resize
from scipy.interpolate import interpn

from ..lddmm._lddmm_utilities import _validate_scalar_to_multi
from ..lddmm._lddmm_utilities import _validate_ndarray
from ..lddmm._lddmm_utilities import _validate_resolution
from ..lddmm._lddmm_utilities import _compute_axes
from ..lddmm._lddmm_utilities import _compute_coords

def _resample(image, resolution, final_resolution, **interpn_kwargs):
    # interpn wrapper, not to be used without achievable final_resolution.

    if not np.all(image.shape * resolution % final_resolution):
        raise ValueError(f"final_resolution must evenly fit into the real shape of image.")

    final_shape = np.array(image.shape * resolution / final_resolution, int)
    
    # Validate inputs.
    image = _validate_ndarray(image)
    resolution = _validate_resolution(resolution, image.ndim)
    desired_resolution = _validate_resolution(desired_resolution, image.ndim)
    
    # Construct arguments for scipy.interpolate.interpn.
    real_axes = _compute_axes(shape=image.shape, resolution=resolution)
    new_real_coords = _compute_coords(shape=final_shape, resolution=final_resolution)

    return interpn(points=real_axes, values=image, xi=new_real_coords, **interpn_kwargs)


def change_resolution_to(image, resolution, desired_resolution, 
pad_to_match_res=True, err_to_higher_res=True, return_final_resolution=False, **resize_kwargs):
    

    # Validate inputs.
    image = _validate_ndarray(image)
    resolution = _validate_resolution(resolution, image.ndim)
    desired_resolution = _validate_resolution(desired_resolution, image.ndim)
    
    # Compute final_shape and final_resolution.

    # Take the exact floating point final_shape that would be necessary to achieve desired_resolution.
    final_shape = image.shape * resolution / desired_resolution

    # Adjust final_shape and compute final_resolution according to specification.
    if pad_to_match_res:
        # Guarantee realization of desired_resolution at the possible expense of maintaining the true shape (shape * resolution).
        # Note: "true shape" implies real-valued shape: the product of image shape and corresponding resolution.
        final_shape = np.ceil(final_shape)
        # Pad image evenly until image.shape * resolution >= final_shape * desired_resolution.
        minimum_image_padding = np.ceil((final_shape * desired_resolution - image.shape * resolution) / resolution)
        pad_width = np.array(list(zip(np.ceil(minimum_image_padding / 2), np.ceil(minimum_image_padding / 2))), int)
        old_true_shape = resolution * image.shape
        new_true_shape = desired_resolution * final_shape
        stat_length = np.maximum(1, np.ceil((desired_resolution - ((new_true_shape - old_true_shape) / 2)) / resolution)).astype(int)
        stat_length = np.broadcast_to(stat_length, pad_width.T.shape).T
        image = np.pad(image, pad_width=pad_width, mode='mean', stat_length=stat_length) # Side effect: breaks alias.
        # final_resolution has been guaranteed to equal desired_resolution.
        final_resolution = desired_resolution
    else:
        # Guarantee the true shape (image.shape * resolution) is maintained at the possible expense of achieving desired_resolution.
        if err_to_higher_res:
            # Round resolution up.
            final_shape = np.ceil(final_shape)
        else:
            # Round resolution down.
            final_shape = np.floor(final_shape)
        # Compute the achieved resultant resolution, or final_resolution.
        final_resolution = image.shape * resolution / final_shape
        # Warn the user if desired_resolution cannot be produced from image and resolution.
        if not np.array_equal(final_shape, image.shape * resolution / desired_resolution): # If desired_resolution != final_resolution.
            warnings.warn(message=f"Could not exactly produce the desired_resolution.\n"
                f"xyz_resolution {xyz_resolution}.\n"
                f"desired_resolution: {desired_resolution}.\n"
                f"final_resolution: {final_resolution}.", category=RuntimeWarning)

    # Note: if the function used for resampling below does not include anti-aliasing in the dase of downsampling, 
    # that ought to be performed here. skimage.transform.resize does perform anti-aliasing by default.

    # Perform resampling.

    resampled_image = resize(image, final_shape, **resize_kwargs)
    return resampled_image


def change_resolution_by(image, scales, resolution=1, 
pad_to_match_res=True, err_to_higher_res=True, return_final_resolution=False, **resize_kwargs):
    

    image = _validate_ndarray(image)
    scales = _validate_scalar_to_multi(scales, size=image.ndim)
    resolution = _validate_resolution(resolution, image.ndim)
    desired_resolution = resolution / scales

    change_resolution_to_kwargs = dict(
        image=image,
        resolution=resolution,
        desired_resolution=desired_resolution,
        pad_to_match_res=pad_to_match_res,
        err_to_higher_res=err_to_higher_res,
        return_final_resolution=return_final_resolution,
        **resize_kwargs
    )

    return change_resolution_to(**change_resolution_to_kwargs)



import pytest

import numpy as np

from ._lddmm_utilities import _validate_scalar_to_multi
from ._lddmm_utilities import _validate_ndarray
from ._lddmm_utilities import _validate_resolution
from ._lddmm_utilities import _compute_axes
from ._lddmm_utilities import _compute_coords
from ._lddmm_utilities import _multiply_coords_by_affine
from ._lddmm_utilities import resample
from ._lddmm_utilities import sinc_resample

"""
Test _validate_scalar_to_multi.
"""

class Test__validate_scalar_to_multi:

    # Test proper use.

    def test_singular(self):
        kwargs = dict(value=1, size=1, dtype=float)
        correct_output = np.array([1], float)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)

    def test_empty(self):
        kwargs = dict(value=1, size=0, dtype=int)
        correct_output = np.array([], int)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)

    def test_scalar_dtype_casting(self):
        kwargs = dict(value=9.5, size=4, dtype=int)
        correct_output = np.full(4, 9, int)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)

    def test_list_value_dtype_casting_to_float(self):
        kwargs = dict(value=[1, 2, 3.5], size=3, dtype=float)
        correct_output = np.array([1, 2, 3.5], float)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)

    def test_list_value_dtype_casting_to_int(self):
        kwargs = dict(value=[1, 2, 3.5], size=3, dtype=int)
        correct_output = np.array([1, 2, 3], int)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)

    def test_tuple_value(self):
        kwargs = dict(value=(1, 2, 3), size=3, dtype=int)
        correct_output = np.array([1, 2, 3], int)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)

    def test_array_value(self):
        kwargs = dict(value=np.array([1, 2, 3], float), size=3, dtype=int)
        correct_output = np.array([1, 2, 3], int)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)

    def test_multi_value_with_size_None(self):
        kwargs = dict(value=np.array([1, 2, 3], float), size=None, dtype=float)
        correct_output = np.array([1, 2, 3], float)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)


    def test_scalar_value_with_size_None(self):
        kwargs = dict(value=1, size=None, dtype=float)
        correct_output = np.array([1], float)
        assert np.array_equal(_validate_scalar_to_multi(**kwargs), correct_output)

    # Test improper use.

    def test_non_int_size(self):
        kwargs = dict(value=[1, 2, 3, 4], size='size: not an int', dtype=float)
        expected_exception = TypeError
        match = "size must be either None or interpretable as an integer."
        with pytest.raises(expected_exception, match=match):
            _validate_scalar_to_multi(**kwargs)

    def test_negative_size(self):
        kwargs = dict(value=[], size=-1, dtype=float)
        expected_exception = ValueError
        match = "size must be non-negative."
        with pytest.raises(expected_exception, match=match):
            _validate_scalar_to_multi(**kwargs)

    def test_value_incompatible_with_size(self):
        kwargs = dict(value=[1, 2, 3, 4], size=3, dtype=int)
        expected_exception = ValueError
        match = "The length of value must either be 1 or it must match size if size is provided."
        with pytest.raises(expected_exception, match=match):
            _validate_scalar_to_multi(**kwargs)

    def test_multi_dimensional_value(self):
        kwargs = dict(value=np.arange(3*4, dtype=int).reshape(3,4), size=3, dtype=float)
        expected_exception = ValueError
        match = "value must not have more than 1 dimension."
        with pytest.raises(expected_exception, match=match):
            _validate_scalar_to_multi(**kwargs)

    def test_list_value_uncastable_to_dtype(self):
        kwargs = dict(value=[1, 2, 'c'], size=3, dtype=int)
        expected_exception = ValueError
        match = "value and dtype are incompatible with one another."
        with pytest.raises(expected_exception, match=match):
            _validate_scalar_to_multi(**kwargs)

    def test_scalar_value_uncastable_to_dtype(self):
        kwargs = dict(value='c', size=3, dtype=int)
        expected_exception = ValueError
        match = "value and dtype are incompatible with one another."
        with pytest.raises(expected_exception, match=match):
            _validate_scalar_to_multi(**kwargs)

"""
Test _validate_ndarray.
"""

class Test__validate_ndarray:

    # Test proper use.

    def test_cast_array_to_dtype(self):
        kwargs = dict(array=np.arange(3, dtype=int), dtype=float)
        correct_output = np.arange(3, dtype=float)
        assert np.array_equal(_validate_ndarray(**kwargs), correct_output)

    def test_array_as_list(self):
        kwargs = dict(array=[[0,1,2], [3,4,5]], dtype=float)
        correct_output = np.arange(2*3, dtype=float).reshape(2,3)
        assert np.array_equal(_validate_ndarray(**kwargs), correct_output)

    def test_scalar_array_upcast_by_minimum_ndim(self):
        kwargs = dict(array=9, minimum_ndim=2)
        correct_output = np.array(9).reshape(1, 1)
        assert np.array_equal(_validate_ndarray(**kwargs), correct_output)

    def test_broadcast_to_shape(self):
        kwargs = dict(array=np.array([0,1,2]), broadcast_to_shape=(2,3))
        correct_output = np.array([[0,1,2], [0,1,2]])
        assert np.array_equal(_validate_ndarray(**kwargs), correct_output)

    def test_reshape_to_shape(self):
        kwargs = dict(array=np.arange(3 * 4).reshape(3, 4), reshape_to_shape=(4, 3))
        correct_output = np.arange(3 * 4).reshape(4, 3)
        assert np.array_equal(_validate_ndarray(**kwargs), correct_output)
    
    def test_wildcard_reshape_to_shape(self):
        kwargs = dict(array=np.arange(3 * 4).reshape(3, 4), reshape_to_shape=(4, -1))
        correct_output = np.arange(3 * 4).reshape(4, 3)
        assert np.array_equal(_validate_ndarray(**kwargs), correct_output)

    def test_wildcard_required_shape(self):
        kwargs = dict(array=np.arange(3 * 4).reshape(3, 4), required_shape=(3, -1))
        correct_output = np.arange(3 * 4).reshape(3, 4)
        assert np.array_equal(_validate_ndarray(**kwargs), correct_output)

    # Test improper use.

    # Validate arguments.

    def test_non_int_minimum_ndim(self):
        kwargs = dict(array=np.arange(3), minimum_ndim=1.5)
        expected_exception = TypeError
        match = "minimum_ndim must be of type int."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

    def test_negative_minimum_ndim(self):
        kwargs = dict(array=np.arange(3), minimum_ndim=-1)
        expected_exception = ValueError
        match = "minimum_ndim must be non-negative."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

    def test_non_int_required_ndim(self):
        kwargs = dict(array=np.arange(3), required_ndim=1.5)
        expected_exception = TypeError
        match = "required_ndim must be either None or of type int."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

    def test_negative_required_ndim(self):
        kwargs = dict(array=np.arange(3), required_ndim=-1)
        expected_exception = ValueError
        match = "required_ndim must be non-negative."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

    def test_non_type_dtype(self):
        kwargs = dict(array=np.arange(3), dtype="not of type type")
        expected_exception = TypeError
        match = "dtype must be either None or a valid type."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)
        
    # Validate array.

    def test_array_type_incompataible_with_dtype(self):
        kwargs = dict(array=np.array(print), dtype=int)
        expected_exception = TypeError
        match = "array is of a type that is incompatible with dtype."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

    def test_array_value_incompatible_with_dtype(self):
        kwargs = dict(array=np.array('string that is not an int'), dtype=int)
        expected_exception = ValueError
        match = "array has a value that is incompatible with dtype."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)
        
    def test_forbid_object_dtype(self):
        kwargs = dict(array=np.array([[], 1]), dtype=None, forbid_object_dtype=True)
        expected_exception = TypeError
        match = "Casting array to a np.ndarray produces an array of dtype object \nwhile forbid_object_dtype == True and dtype != object."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

    def test_required_ndim(self):
        kwargs = dict(array=np.arange(3), required_ndim=2)
        expected_exception = ValueError
        match = "If required_ndim is not None, array.ndim must be made to equal it."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

    def test_incompaatible_required_shape(self):
        kwargs = dict(array=np.arange(3 * 4).reshape(3, 4), required_shape=(3, 5))
        expected_exception = ValueError
        match = "array is incompatible with required_shape."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

    def test_compatible_but_not_matched_required_shape(self):
        kwargs = dict(array=np.arange(3 * 4).reshape(3, 4), required_shape=(4, -1))
        expected_exception = ValueError
        match = "array is compatible with required_shape but does not match required_shape."
        with pytest.raises(expected_exception, match=match):
            _validate_ndarray(**kwargs)

"""
Test _validate_resolution.
"""

class Test__validate_resolution:

    # Test proper use.
    
    def test_scalar_resolution_1D_ndim(self):
        kwargs = dict(resolution=2, ndim=1, dtype=float)
        correct_output = np.full(1, 2, float)
        assert np.array_equal(_validate_resolution(**kwargs), correct_output)

    def test_scalar_resolution_4D_ndim(self):
        kwargs = dict(resolution=1.5, ndim=4, dtype=float)
        correct_output = np.full(4, 1.5, float)
        assert np.array_equal(_validate_resolution(**kwargs), correct_output)

    def test_array_resolution_3D_ndim(self):
        kwargs = dict(resolution=np.ones(3, int), ndim=3, dtype=float)
        correct_output = np.ones(3, float)
        assert np.array_equal(_validate_resolution(**kwargs), correct_output)

    def test_list_resolution_2D_ndim(self):
        kwargs = dict(resolution=[3, 4], ndim=2, dtype=float)
        correct_output = np.array([3, 4], float)
        assert np.array_equal(_validate_resolution(**kwargs), correct_output)

        # Test improper use.

    def test_negative_resolution(self):
        kwargs = dict(resolution=[3, -4], ndim=2, dtype=float)
        expected_exception = ValueError
        match = "All elements of resolution must be positive."
        with pytest.raises(expected_exception, match=match):
            _validate_resolution(**kwargs)

    def test_zero_resolution(self):
        kwargs = dict(resolution=[3, 0], ndim=2, dtype=float)
        expected_exception = ValueError
        match = "All elements of resolution must be positive."
        with pytest.raises(expected_exception, match=match):
            _validate_resolution(**kwargs)

"""
Test _compute_axes.
"""

class Test__compute_axes:

    # Test proper use.

    # _compute_axes produces a list with a np.ndarray for each element in shape.

    def test_zero_and_one_in_shape(self):
        kwargs = dict(shape=(0, 1, 2), resolution=1, origin='center')
        correct_output = [np.arange(dim_size) * dim_res - np.mean(np.arange(dim_size) * dim_res) 
            for dim_size, dim_res in zip((0, 1, 2), (1, 1, 1))]
        for dim, coord in enumerate(_compute_axes(**kwargs)):
            assert np.array_equal(coord, correct_output[dim])

    def test_decimal_resolution(self):
        kwargs = dict(shape=(1, 2, 3, 4), resolution=1.5, origin='center')
        correct_output = [np.arange(dim_size) * dim_res - np.mean(np.arange(dim_size) * dim_res) 
            for dim_size, dim_res in zip((1, 2, 3, 4), (1.5, 1.5, 1.5, 1.5))]
        for dim, coord in enumerate(_compute_axes(**kwargs)):
            assert np.array_equal(coord, correct_output[dim])

    def test_anisotropic_resolution(self):
        kwargs = dict(shape=(2, 3, 4), resolution=[1, 1.5, 2], origin='center')
        correct_output = [np.arange(dim_size) * dim_res - np.mean(np.arange(dim_size) * dim_res) 
            for dim_size, dim_res in zip((2, 3, 4), (1, 1.5, 2))]
        for dim, coord in enumerate(_compute_axes(**kwargs)):
            assert np.array_equal(coord, correct_output[dim])

    def test_1D_shape(self):
        kwargs = dict(shape=5, resolution=1, origin='center')
        correct_output = [np.arange(dim_size) * dim_res - np.mean(np.arange(dim_size) * dim_res) 
            for dim_size, dim_res in zip((5,), (1,))]
        for dim, coord in enumerate(_compute_axes(**kwargs)):
            assert np.array_equal(coord, correct_output[dim])

    def test_zero_origin(self):
        kwargs = dict(shape=5, resolution=1, origin='zero')
        correct_output = [np.arange(dim_size) * dim_res
            for dim_size, dim_res in zip((5,), (1,))]
        for dim, coord in enumerate(_compute_axes(**kwargs)):
            assert np.array_equal(coord, correct_output[dim])

"""
Test _compute_coords.
"""

class Test__compute_coords:

    # Test proper use.

    def test_1D_shape_center_origin(self):
        kwargs = dict(shape=5, resolution=1, origin='center')
        correct_output = np.array([[-2], [-1], [0], [1], [2]])
        assert np.array_equal(_compute_coords(**kwargs), correct_output)

    def test_2D_shape_zero_origin(self):
        kwargs = dict(shape=(3,4), resolution=1, origin='zero')
        correct_output = np.array([[[0,0], [0,1], [0,2], [0,3]], [[1,0], [1,1], [1,2], [1,3]], [[2,0], [2,1], [2,2], [2,3]]])
        assert np.array_equal(_compute_coords(**kwargs), correct_output)

"""
Test _multiply_coords_by_affine.
"""

class Test__multiply_coords_by_affine:

    # Test proper use.

    def test_identity_affine_3D(self):
        affine = np.eye(4) + np.append(np.arange(3*4).reshape(3,4), np.zeros((1,4)), 0)**2
        array = _compute_coords((3,4,5), 1)
        result = _multiply_coords_by_affine(affine, array)
        arrays = []
        for dim in range(3):
            arrays.append(np.sum(affine[dim, :-1] * array, axis=-1) + affine[dim, -1])
        expected = np.stack(arrays=arrays, axis=-1)
        assert np.array_equal(result, expected)

    # Test improper use.

    def test_3D_affine_array(self):
        affine = np.eye(4)[None]
        array = np.arange(4*5*6*3).reshape(4,5,6,3)
        expected_exception = ValueError
        match = "affine must be a 2-dimensional matrix."
        with pytest.raises(expected_exception, match=match):
            _multiply_coords_by_affine(affine, array)

    def test_non_square_affine(self):
        affine = np.eye(4)[:3]
        array = np.arange(4*5*6*3).reshape(4,5,6,3)
        expected_exception = ValueError
        match = "affine must be a square matrix."
        with pytest.raises(expected_exception, match=match):
            _multiply_coords_by_affine(affine, array)

    def test_incompatible_affine_and_array(self):
        affine = np.eye(4)
        array = np.arange(4*5*6*2).reshape(4,5,6,2)
        expected_exception = ValueError
        match = "array is incompatible with affine. The length of the last dimension of array should be 1 less than the length of affine."
        with pytest.raises(expected_exception, match=match):
            _multiply_coords_by_affine(affine, array)

    # Verify warning on violation of homogenous coordinates in affine.
    
    def test_homogenous_coordinates_violation_warning(self):
        affine = np.ones((4,4))
        array = np.arange(4*5*6*3).reshape(4,5,6,3)
        expected_warning = RuntimeWarning
        match = "affine is not in homogenous coordinates.\naffine\[-1] should be zeros with a 1 on the right."
        with pytest.warns(expected_warning, match=match):
            _multiply_coords_by_affine(affine, array)
 
"""
Test resample.
"""

class Test_resample:

    # Test proper use.

    def test_zero_origin_upsample(self):
        kwargs = dict(
            image=np.array([
                [0, 1, 2], 
                [3, 4, 5], 
            ]), 
            new_resolution=1/2, 
            old_resolution=1, 
            err_to_larger=True, 
            extrapolation_fill_value=None, 
            origin='zero', 
            method='linear', 
            anti_aliasing=False, 
        )
        correct_output = np.array([
            [0.0, 0.5, 1.0, 1.5, 2.0, 2.5], 
            [1.5, 2.0, 2.5, 3.0, 3.5, 4.0], 
            [3.0, 3.5, 4.0, 4.5, 5.0, 5.5], 
            [4.5, 5.0, 5.5, 6.0, 6.5, 7.0], 
        ])
        assert np.array_equal(resample(**kwargs), correct_output)

    def test_center_origin_upsample(self):
        kwargs = dict(
            image=np.array([
                [0, 1, 2], 
                [3, 4, 5], 
            ]), 
            new_resolution=1/2, 
            old_resolution=1, 
            err_to_larger=True, 
            extrapolation_fill_value=None, 
            origin='center', 
            method='linear', 
            anti_aliasing=False, 
        )
        correct_output = np.array([
            [-1.0, -0.5,  0.0,  0.5,  1.0,  1.5], 
            [ 0.5,  1.0,  1.5,  2.0,  2.5,  3.0], 
            [ 2.0,  2.5,  3.0,  3.5,  4.0,  4.5], 
            [ 3.5,  4.0,  4.5,  5.0,  5.5,  6.0], 
        ])
        assert np.array_equal(resample(**kwargs), correct_output)

    def test_zero_origin_downsample(self):
        kwargs = dict(
            image=np.array([
                [ 0,  1,  2,  3,  4], 
                [ 5,  6,  7,  8,  9], 
                [10, 11, 12, 13, 14], 
                [15, 16, 17, 18, 19], 
            ]), 
            new_resolution=2, 
            old_resolution=1, 
            err_to_larger=True, 
            extrapolation_fill_value=None, 
            origin='zero', 
            method='linear', 
            anti_aliasing=False, 
        )
        correct_output = np.array([
            [ 0,  2,  4], 
            [10, 12, 14], 
        ])
        assert np.array_equal(resample(**kwargs), correct_output)

    def test_center_origin_downsample(self):
        kwargs = dict(
            image=np.array([
                [ 0,  1,  2,  3,  4], 
                [ 5,  6,  7,  8,  9], 
                [10, 11, 12, 13, 14], 
                [15, 16, 17, 18, 19], 
            ]), 
            new_resolution=2, 
            old_resolution=1, 
            err_to_larger=True, 
            extrapolation_fill_value=None, 
            origin='center', 
            method='linear', 
            anti_aliasing=False, 
        )
        correct_output = np.array([
            [ 2.5,  4.5,  6.5],
            [12.5, 14.5, 16.5], 
        ])
        assert np.array_equal(resample(**kwargs), correct_output)

    def test_zero_origin_joint_upsample_and_downsample(self):
        kwargs = dict(
            image=np.array([
                [0, 1, 2, 3, 4], 
                [5, 6, 7, 8, 9], 
            ]), 
            new_resolution=[1/2, 2], 
            old_resolution=1, 
            err_to_larger=True, 
            extrapolation_fill_value=None, 
            origin='zero', 
            method='linear', 
            anti_aliasing=False, 
        )
        correct_output = np.array([
            [   0,    2,    4], 
            [ 2.5,  4.5,  6.5], 
            [ 5.0,  7.0,  9.0], 
            [ 7.5,  9.5, 11.5], 
        ])
        assert np.array_equal(resample(**kwargs), correct_output)

    def test_center_origin_joint_upsample_and_downsample(self):
        kwargs = dict(
            image=np.array([
                [0, 1, 2, 3, 4], 
                [5, 6, 7, 8, 9], 
            ]), 
            new_resolution=[1/2, 2], 
            old_resolution=1, 
            err_to_larger=True, 
            extrapolation_fill_value=None, 
            origin='center', 
            method='linear', 
            anti_aliasing=False, 
        )
        correct_output = np.array([
            [-1.25,  0.75,  2.75], 
            [ 1.25,  3.25,  5.25], 
            [ 3.75,  5.75,  7.75], 
            [ 6.25,  8.25, 10.25], 
        ])
        assert np.array_equal(resample(**kwargs), correct_output)

    def test_center_origin_joint_upsample_and_downsample_with_err_to_larger_False(self):
        kwargs = dict(
            image=np.array([
                [ 0,  1,  2,  3,  4], 
                [ 5,  6,  7,  8,  9], 
                [10, 11, 12, 13, 14], 
            ]), 
            new_resolution=[1/2, 2], 
            old_resolution=1, 
            err_to_larger=False, 
            extrapolation_fill_value=None, 
            origin='zero', 
            method='linear', 
            anti_aliasing=False, 
        )
        correct_output = np.array([
            [   0,    2], 
            [ 2.5,  4.5], 
            [ 5.0,  7.0], 
            [ 7.5,  9.5], 
            [10.0, 12.0], 
            [12.5, 14.5], 
        ])
        assert np.array_equal(resample(**kwargs), correct_output)

"""
Test sinc_resample.
"""
# @pytest.mark.skip
class Test_sinc_resample:

    # Test shape of an upsample.
    def test_upsample_shape(self):
        kwargs = dict(
            array=np.arange(10*20*30).reshape(10,20,30),
            new_shape=(12, 45, 30),
        )
        output = sinc_resample(**kwargs)
        assert np.array_equal(output.shape, kwargs['new_shape'])

    # Test shape of a downsample.
    def test_downsample_shape(self):
        kwargs = dict(
            array=np.arange(10*20*30).reshape(10,20,30),
            new_shape=(8, 20, 14),
        )
        output = sinc_resample(**kwargs)
        assert np.array_equal(output.shape, kwargs['new_shape'])

    # Test shape of a general resample.
    def test_joint_upsample_and_downsample_shape(self):
        kwargs = dict(
            array=np.arange(10*20*30).reshape(10,20,30),
            new_shape=(10, 23, 27),
        )
        output = sinc_resample(**kwargs)
        assert np.array_equal(output.shape, kwargs['new_shape'])

    # Test preservation of the trends in a general resample.
    def test_preserves_trends(self):
        kwargs = dict(
            array=np.array([
                [0, 1, 2, 3, 4, 3, 2, 1, 0],
                [2, 3, 4, 5, 6, 5, 4, 3, 2],
                [0, 1, 2, 3, 4, 3, 2, 1, 0],
            ]), # Shape: (3, 9).
            new_shape=(5, 7),
        )
        output = sinc_resample(**kwargs)
        # Check that the first half of rows are ascending.
        for row in range((output.shape[0] - 1) // 2):
            assert np.all(output[row, :] <= output[row + 1, :])
        # Check that the second half of rows are descending.
        for row in range(output.shape[0] // 2, output.shape[0] - 1):
            assert np.all(output[row, :] >= output[row + 1, :])
        # Check that the first half of columns are ascending.
        for col in range((output.shape[1] - 1) // 2):
            assert np.all(output[:, col] <= output[:, col + 1])
        # Check that the second half of columns are descending.
        for col in range(output.shape[1] // 2, output.shape[1] - 1):
            assert np.all(output[:, col] >= output[:, col + 1])
        

"""
Perform tests.
"""

if __name__ == "__main__":
    test__validate_scalar_to_multi()
    test__validate_ndarray()
    test__validate_resolution()
    test__compute_axes()
    test__compute_coords()
    test__multiply_coords_by_affine()
    test_resample()
    test_sinc_resample()



from .normalization import cast_to_typed_array
from .normalization import normalize_by_MAD
from .normalization import center_to_mean
from .normalization import pad

from .bias_and_artifact_correction import correct_bias_field
from .bias_and_artifact_correction import remove_grid_artifact

from .initialization import locally_rotate_velocity_fields
from .initialization import locally_translate_velocity_fields


# TODO: update preprocessing_functions, include resample.
preprocessing_functions = [
    # from .normalization:
    'cast_to_typed_array',
    'normalize_by_MAD',
    'center_to_mean',
    'pad',
    # from .bias_and_artifact_correction:
    'correct_bias_field',
    'remove_grid_artifact',
    # from .initialization:
    'locally_rotate_velocity_fields',
    'locally_translate_velocity_fields',
    ]

"""
Preprocessing Pipeline
"""

from collections.abc import Iterable

import numpy as np

from ..lddmm._lddmm_utilities import _validate_ndarray

def preprocess(data, processes):
    """
    Perform each preprocessing function in processes, in the order listed, 
    on data if it is an array, or on each element in data if it is a list of arrays.
    
    Args:
        data (np.ndarray, list): The array or list of arrays to be preprocessed.
        processes (list): The list of strings, each corresponding to the name of a preprocessing function.
        process_kwargs (seq, optional): A sequence of dictionaries containing kwargs for each element of processes. Defaults to None.
    
    Raises:
        TypeError: Raised if data is a list whose elements are not all of type np.ndarray.
        TypeError: Raised if data is neither a np.ndarray or a list of np.ndarrays.
        TypeError: Raised if an element of processes is neither a single string nor an iterable.
        ValueError: Raised if an element of processes is an iterable but not of length 2.
        ValueError: Raised if an alement of processes is a 2-element iterable whose first element is not a string.
        ValueError: Raised if an alement of processes is a 2-element iterable whose second element is not a dictionary.
        TypeError: Raised if an element of processes includes a dictionary with a key that is not a string.
        ValueError: Raised if any element of processes is not a recognized preprocessing function.
    
    Returns:
        np.ndarray, list: A copy of data after having each function in processes applied.
    """

    # Check data form to match output.
    data_given_as_list = isinstance(data, list)

    # Validate inputs.

    # Verify data.
    if isinstance(data, list):
        if not all(isinstance(datum, np.ndarray) for datum in data):
            raise TypeError(f"If data is a list, all elements must be np.ndarrays.\n"
                f"type(data[0]): {type(data[0])}.")
    elif isinstance(data, np.ndarray):
        data = [data]
    else:
        # data is neither a list nor a np.ndarray.
        raise TypeError(f"data must be a np.ndarray or a list of np.ndarrays.")
    
    # Validate processes.
    processes = _validate_ndarray(processes, required_ndim=1)
    for process_index, process in enumerate(processes):
        # Check whether this process is just a single string.
        if isinstance(process, str):
            continue
        # This process is not a single string.
        # Check whether it is an Iterable.
        if not isinstance(process, Iterable):
            raise TypeError(f"If an element of processes is not a single string then it must be an iterable.\n"
                            f"type(processes[{process_index}]: {type(process)}.")
        # This process is an Iterable.
        # Check whether this process has length 2.
        if len(process) != 2:
            raise ValueError(f"If an element of processes is not a single string then it must be a 2-element iterable.\n"
                             f"type(processes[{process_index}]): {type(process)}.\n"
                             f"len(processes[{process_index}]): {len(process)}.")
        # This process is an Iteraable of length 2.
        # Check whether the first element of this process is a string.
        if not isinstance(process[0], str):
            raise ValueError(f"If an element of processes is a 2-element iterable, the first element must be of type str.\n"
                             f"type(processes[{process_index}][0]: {type(process[0])}.")
        # Check whether the second element of this process is a dict.
        if not isinstance(process[1], dict):
            raise ValueError(f"If an element of processes is a 2-element iterable, the second element must be of type dict.\n"
                                f"type(processes[{process_index}][1]: {type(process[1])}.")
        # Check whether all keys of the process[1] dictionary are strings.
        for key_index, key in enumerate(process[1].keys()):
            if not isinstance(key, str):
                raise TypeError(f"Each dictionary of kwargs must all keys be of type str.\n"
                                f"type(processes[{process_index}][1].keys()[{key_index}]): {type(key)}.")

    # Process each np.ndarray.
    # If data was passed in as a single np.ndarray, 
    # then data is now a 1-element list containing that np.ndarray: [data].
    for data_index, datum in enumerate(data):
        for process_index, process in enumerate(processes):
            # process is either a string or a 2-element sequence whose first element is a string and whose second element is a dictionary of kwargs for that process.
            if isinstance(process, str):
                process, process_kwargs = process, dict()
            else:
                process, process_kwargs = process
            # process is a string indicating a preprocessing function, process_kwargs is a dictionary of kwargs to be passed to that process.
            if process in preprocessing_functions:
                datum = eval(f"{process}(datum, **process_kwargs)")
            else:
                raise ValueError(f"Process {process} not recognized.\n"
                    f"Recognized processes: {preprocessing_functions}.")
        data[data_index] = datum

    # Return in a form appropriate to what was passed in, 
    # i.e. list in, list out, np.ndarray in, np.ndarray out.
    return data if data_given_as_list else data[0]



import pytest
from itertools import product

import numpy as np
from scipy.linalg import inv
from scipy.ndimage import rotate

from . import _lddmm_utilities

from ._lddmm import generate_position_field
from ._lddmm import _transform_image
from ._lddmm import lddmm_transform_image
from ._lddmm import lddmm_register
from ._lddmm import _transform_points
from ._lddmm import lddmm_transform_points

"""
Test generate_position_field.
"""

@pytest.mark.parametrize('deform_to', ['template', 'target'])
class Test_generate_position_field:

    def test_identity_affine_identity_velocity_fields(self, deform_to):

        num_timesteps = 10

        template_shape = (3,4,5)
        template_resolution = 1
        target_shape = (2,4,6)
        target_resolution = 1
        velocity_fields = np.zeros((*template_shape, num_timesteps, len(template_shape)))
        velocity_field_resolution = 1
        affine = np.eye(4)

        if deform_to == 'template':
            expected_output = _lddmm_utilities._compute_coords(template_shape, template_resolution)
        elif deform_to == 'target':
            expected_output = _lddmm_utilities._compute_coords(target_shape, target_resolution)

        position_field = generate_position_field(
            affine=affine,
            velocity_fields=velocity_fields,
            velocity_field_resolution=velocity_field_resolution,
            template_shape=template_shape,
            template_resolution=template_resolution,
            target_shape=target_shape,
            target_resolution=target_resolution,
            deform_to=deform_to,
        )

        assert np.array_equal(position_field, expected_output)

    def test_identity_affine_constant_velocity_fields(self, deform_to):

        num_timesteps = 10

        template_shape = (3,4,5)
        template_resolution = 1
        target_shape = (2,4,6)
        target_resolution = 1
        velocity_fields = np.ones((*template_shape, num_timesteps, len(template_shape)))
        velocity_field_resolution = 1
        affine = np.eye(4)

        if deform_to == 'template':
            expected_output = _lddmm_utilities._compute_coords(template_shape, template_resolution) + 1
        elif deform_to == 'target':
            expected_output = _lddmm_utilities._compute_coords(target_shape, target_resolution) - 1

        position_field = generate_position_field(
            affine=affine,
            velocity_fields=velocity_fields,
            velocity_field_resolution=velocity_field_resolution,
            template_shape=template_shape,
            template_resolution=template_resolution,
            target_shape=target_shape,
            target_resolution=target_resolution,
            deform_to=deform_to,
        )

        assert np.allclose(position_field, expected_output)

    def test_rotational_affine_identity_velocity_fields(self, deform_to):

        num_timesteps = 10

        template_shape = (3,4,5)
        template_resolution = 1
        target_shape = (2,4,6)
        target_resolution = 1
        velocity_fields = np.zeros((*template_shape, num_timesteps, len(template_shape)))
        velocity_field_resolution = 1
        # Indicates a 90 degree rotation to the right.
        affine = np.array([
            [0,1,0,0], 
            [-1,0,0,0], 
            [0,0,1,0], 
            [0,0,0,1], 
        ])

        if deform_to == 'template':
            expected_output = _lddmm_utilities._multiply_coords_by_affine(affine, 
                _lddmm_utilities._compute_coords(template_shape, template_resolution))
        elif deform_to == 'target':
            expected_output = _lddmm_utilities._multiply_coords_by_affine(inv(affine), 
                _lddmm_utilities._compute_coords(target_shape, target_resolution))

        position_field = generate_position_field(
            affine=affine,
            velocity_fields=velocity_fields,
            velocity_field_resolution=velocity_field_resolution,
            template_shape=template_shape,
            template_resolution=template_resolution,
            target_shape=target_shape,
            target_resolution=target_resolution,
            deform_to=deform_to,
        )
        
        assert np.allclose(position_field, expected_output)

"""
Test _transform_image.
"""

class Test__transform_image:
    
    def test_identity_position_field_equal_output_resolution(self):

        subject = np.arange(3*4).reshape(3,4)
        subject_resolution = 1
        output_resolution = 1
        output_shape=None
        position_field_resolution = subject_resolution
        position_field = _lddmm_utilities._compute_coords(subject.shape, position_field_resolution)

        deformed_subject = _transform_image(
            subject=subject,
            subject_resolution=subject_resolution,
            output_resolution=output_resolution,
            output_shape=output_shape,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )
        expected_output = subject
        assert np.allclose(deformed_subject, expected_output)

    def test_identity_position_field_different_output_resolution(self):

        subject = np.array([
            [0,0,0,0,0,0,0,0],
            [0,0,0,0,0,0,0,0],
            [0,0,1,1,1,1,0,0],
            [0,0,1,1,1,1,0,0],
            [0,0,1,1,1,1,0,0],
            [0,0,1,1,1,1,0,0],
            [0,0,0,0,0,0,0,0],
            [0,0,0,0,0,0,0,0],
        ])
        subject_resolution = 1
        output_resolution = 2
        output_shape = None
        position_field_resolution = subject_resolution
        position_field = _lddmm_utilities._compute_coords(subject.shape, position_field_resolution)

        deformed_subject = _transform_image(
            subject=subject,
            subject_resolution=subject_resolution,
            output_resolution=output_resolution,
            output_shape=output_shape,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )
        expected_output = np.array([
            [0,0,0,0],
            [0,1,1,0],
            [0,1,1,0],
            [0,0,0,0],
        ])
        assert np.allclose(deformed_subject, expected_output)
    
    def test_constant_position_field_trivial_extrapolation(self):

        # Note: applying a leftward shift to the position_field is done by subtracting 1 from the appropriate dimension.
        # The corresponding effect on the deformed_subject is a shift to the right.

        subject = np.array([
            [0,0,0,0,0,0,0,0],
            [0,0,0,0,0,0,0,0],
            [0,0,1,1,1,1,0,0],
            [0,0,1,1,1,1,0,0],
            [0,0,1,1,1,1,0,0],
            [0,0,1,1,1,1,0,0],
            [0,0,0,0,0,0,0,0],
            [0,0,0,0,0,0,0,0],
        ])
        subject_resolution = 1
        output_resolution = 1
        output_shape = None
        position_field_resolution = subject_resolution
        position_field = _lddmm_utilities._compute_coords(subject.shape, position_field_resolution) + [0, -1] # Shift to the left by 1.

        deformed_subject = _transform_image(
            subject=subject,
            subject_resolution=subject_resolution,
            output_resolution=output_resolution,
            output_shape=output_shape,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )
        expected_output = np.array([
            [0,0,0,0,0,0,0,0],
            [0,0,0,0,0,0,0,0],
            [0,0,0,1,1,1,1,0],
            [0,0,0,1,1,1,1,0],
            [0,0,0,1,1,1,1,0],
            [0,0,0,1,1,1,1,0],
            [0,0,0,0,0,0,0,0],
            [0,0,0,0,0,0,0,0],
        ])

        assert np.allclose(deformed_subject, expected_output)

    def test_constant_position_field_linear_extrapolation(self):
    
        # Idiosyncratic extrapolation behavior is demonstrated with a nonzero gradient at the extrapolated edge.

        subject = np.array([
            [0,0,0,0],
            [0,1,1,0],
            [0,1,1,0],
            [0,0,0,0],
        ])
        subject_resolution = 1
        output_resolution = 1
        output_shape = None
        position_field_resolution = subject_resolution
        position_field = _lddmm_utilities._compute_coords(subject.shape, position_field_resolution) + [0, -1] # Shift to the left by 1.

        deformed_subject = _transform_image(
            subject=subject,
            subject_resolution=subject_resolution,
            output_resolution=output_resolution,
            output_shape=output_shape,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )
        expected_output = np.array([
            [0,0,0,0],
            [-1,0,1,1],
            [-1,0,1,1],
            [0,0,0,0],
        ])

        assert np.allclose(deformed_subject, expected_output)

    def test_rotational_position_field(self):

        # Note: applying an affine indicating a clockwise-rotation to a position_field produces a position _ield rotated counter-clockwise.
        # The corresponding effect on the deformed_subject is a counter-clockwise rotation.

        subject = np.array([
            [0,1,0,0],
            [0,1,0,0],
            [0,1,0,0],
            [0,1,1,1],
        ])
        subject_resolution = 1
        output_resolution = 1
        output_shape = None
        position_field_resolution = subject_resolution
        # Indicates a 90 degree rotation to the right.
        affine = np.array([
            [0,1,0],
            [-1,0,0],
            [0,0,1],
        ])
        position_field = _lddmm_utilities._multiply_coords_by_affine(affine, 
            _lddmm_utilities._compute_coords(subject.shape, position_field_resolution))

        deformed_subject = _transform_image(
            subject=subject,
            subject_resolution=subject_resolution,
            output_resolution=output_resolution,
            output_shape=output_shape,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )
        expected_output = np.array([
            [0,0,0,1],
            [0,0,0,1],
            [1,1,1,1],
            [0,0,0,0],
        ])
        
        assert np.allclose(deformed_subject, expected_output)

"""
Test lddmm_transform_image.
"""

@pytest.mark.parametrize('deform_to', ['template', 'target'])
class Test_lddmm_transform_image:

    def test_identity_position_fields(self, deform_to):
    
        subject = np.array([
            [0,0,0,0],
            [0,1,1,0],
            [0,1,1,0],
            [0,0,0,0],
        ])
        subject_resolution = 1
        output_resolution = None
        output_shape = None
        template_shape = (3,4)
        template_resolution = 1
        target_shape = (2,5)
        target_resolution = 1
        extrapolation_fill_value = np.quantile(subject, 10**-subject.ndim)

        affine_phi = _lddmm_utilities._compute_coords(template_shape, template_resolution)
        phi_inv_affine_inv = _lddmm_utilities._compute_coords(target_shape, target_resolution)

        expected_output = _transform_image(
            subject,
            subject_resolution,
            output_resolution,
            output_shape,
            position_field=affine_phi if deform_to == 'template' else phi_inv_affine_inv,
            position_field_resolution=template_resolution if deform_to == 'template' else target_resolution,
            extrapolation_fill_value=extrapolation_fill_value,
        )

        deformed_subject = lddmm_transform_image(
            subject=subject, subject_resolution=subject_resolution, 
            output_resolution=output_resolution, deform_to=deform_to, 
            extrapolation_fill_value=extrapolation_fill_value,
            affine_phi=affine_phi, phi_inv_affine_inv=phi_inv_affine_inv, 
            template_resolution=template_resolution, target_resolution=target_resolution, 
        )

        assert np.array_equal(deformed_subject, expected_output)

"""
Test lddmm_register.
"""
#TODO: Verify that these warnings aren't a problem.
# @pytest.mark.filterwarnings('ignore:Ill-conditioned matrix')
class Test_lddmm_register:

    def _test_lddmm_register(self, rtol=0, atol=1-1e-9, **lddmm_register_kwargs):
        """A helper method for this class to verify registrations once they are computed."""

        reg_output = lddmm_register(**lddmm_register_kwargs)

        template = lddmm_register_kwargs['template']
        target = lddmm_register_kwargs['target']
        template_resolution = lddmm_register_kwargs['template_resolution'] if 'template_resolution' in lddmm_register_kwargs.keys() else 1
        target_resolution = lddmm_register_kwargs['target_resolution'] if 'target_resolution' in lddmm_register_kwargs.keys() else 1

        deformed_target = lddmm_transform_image(
            subject=target, 
            subject_resolution=target_resolution, 
            deform_to='template', 
            **reg_output,
        )

        deformed_template = lddmm_transform_image(
            subject=template, 
            subject_resolution=template_resolution, 
            deform_to='target', 
            **reg_output,
        )

        assert np.allclose(deformed_template, target, rtol=rtol, atol=atol)
        assert np.allclose(deformed_target, template, rtol=rtol, atol=atol)

    def test_2D_identity_registration(self):

        ndims = 2
        radius = 3
        zero_space = 2
        center_pixel = True
        template = np.zeros(tuple([(radius + zero_space) * 2 + center_pixel] * ndims))
        for indices in product(*map(range, template.shape)):
            indices = np.array(indices)
            if np.sqrt(np.sum((indices - (radius + zero_space) + (not center_pixel) / 2)**2)) <= radius:
                template[tuple(indices)] = 1
        target = np.copy(template)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_iterations=1,
        )

        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_3D_identity_registration(self):

        ndims = 3
        radius = 3
        zero_space = 2
        center_pixel = True
        template = np.zeros(tuple([(radius + zero_space) * 2 + center_pixel] * ndims))
        for indices in product(*map(range, template.shape)):
            indices = np.array(indices)
            if np.sqrt(np.sum((indices - (radius + zero_space) + (not center_pixel) / 2)**2)) <= radius:
                template[tuple(indices)] = 1
        target = np.copy(template)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_iterations=1,
        )

        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_4D_identity_registration(self):

        ndims = 4
        radius = 3
        zero_space = 2
        center_pixel = True
        template = np.zeros(tuple([(radius + zero_space) * 2 + center_pixel] * ndims))
        for indices in product(*map(range, template.shape)):
            indices = np.array(indices)
            if np.sqrt(np.sum((indices - (radius + zero_space) + (not center_pixel) / 2)**2)) <= radius:
                template[tuple(indices)] = 1
        target = np.copy(template)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_iterations=1,
        )

        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_identity_disk_to_disk_registration(self):

        template = np.array([[(col-4)**2 + (row-4)**2 <= 4**2 for col in range(9)] for row in range(9)], int)
        target = np.copy(template)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_affine_only_iterations=0,
        )

        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_rigid_affine_only_ellipsoid_to_ellipsoid_registration(self):

        # template (before padding) has shape (21, 29) and semi-radii 4 and 10.
        template = np.array([[(col-14)**2/10**2 + (row-8)**2/4**2 <= 1 for col in range(29)] for row in range(17)], int)
        # templata and target are opposite rotations of an unrotated ellipsoid for symmetry.
        target = rotate(template, 45/2)
        template = rotate(template, -45/2)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_iterations=50,
            num_affine_only_iterations=50,
            num_rigid_affine_iterations=50,
        )

        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_non_rigid_affine_only_ellipsoid_to_ellipsoid_registration(self):

        # template (before padding) has shape (21, 29) and semi-radii 6 and 10.
        template = np.array([[(col-14)**2/10**2 + (row-10)**2/6**2 <= 1 for col in range(29)] for row in range(21)], int)
        # target is a rotation of template.
        target = rotate(template, 30)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_iterations=50,
            num_affine_only_iterations=50,
            num_rigid_affine_iterations=0,
        )

        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_partially_rigid_affine_only_ellipsoid_to_ellipsoid_registration(self):

        # template (before padding) has shape (21, 29) and semi-radii 6 and 10.
        template = np.array([[(col-14)**2/10**2 + (row-10)**2/6**2 <= 1 for col in range(29)] for row in range(21)], int)
        # target is a rotation of template.
        target = rotate(template, 30)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_iterations=100,
            num_affine_only_iterations=100,
            num_rigid_affine_iterations=50,
        )

        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_deformative_only_disk_to_ellipsoid_registration(self):

        # template has shape (25, 25) and radius 8.
        template = np.array([[(col-12)**2 + (row-12)**2 <= 8**2 for col in range(25)] for row in range(25)], int)
        # target has shape (21, 29) and semi-radii 6 and 10.
        target = np.array([[(col-14)**2/10**2 + (row-10)**2/6**2 <= 1 for col in range(29)] for row in range(21)], int)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_iterations=150,
            num_affine_only_iterations=0,
            affine_stepsize=0,
            deformative_stepsize=0.5,
        )
        
        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_general_ellipsoid_to_ellipsoid_registration(self):

        # target has shape (21, 29) and semi-radii 6 and 10.
        template = np.array([[(col-14)**2/10**2 + (row-10)**2/6**2 <= 1 for col in range(29)] for row in range(21)], int)
        # target has shape (21, 29) and semi-radii 6 and 10.
        target = rotate(template, 30)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            deformative_stepsize=0.5,
        )

        self._test_lddmm_register(**lddmm_register_kwargs)

    def test_identity_multiscale_registration(self):

        template = np.zeros((20,30))
        # target has shape (21, 29) and semi-radii 6 and 10.
        template = np.array([[(col-14)**2/12**2 + (row-10)**2/8**2 <= 1 for col in range(29)] for row in range(21)])
        target = np.copy(template)

        lddmm_register_kwargs = dict(
            template=template,
            target=target,
            num_iterations=1,
            # num_affine_only_iterations=[1, 0, 0, 0],
            # num_rigid_affine_iterations=0,
            multiscales=[5, (2,3), [3,2], 1],
        )

        self._test_lddmm_register(**lddmm_register_kwargs)


"""
Test _transform_points.
"""

class Test__transform_points:

    def test_identity_position_field(self):

        subject = np.arange(3*4).reshape(3,4)
        subject_resolution = 1
        position_field_resolution = subject_resolution
        position_field = _lddmm_utilities._compute_coords(subject.shape, position_field_resolution)
        points = _lddmm_utilities._compute_coords(subject.shape, position_field_resolution)

        transformed_points = _transform_points(
            points=points,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )
        expected_output = points
        assert np.array_equal(transformed_points, expected_output)
    
    def test_constant_position_field(self):

        # Note: applying a leftward shift to the position_field is done by subtracting 1 from the appropriate dimension.
        # The corresponding effect on a deformed image is a shift to the right.

        subject = np.array([
            [0,1,3],
            [4,5,6],
            [7,8,9],
        ])
        subject_resolution = 1
        position_field_resolution = subject_resolution
        position_field = _lddmm_utilities._compute_coords(subject.shape, position_field_resolution) + [0, -1] # Shift to the left by 1.
        # The right column.
        points = np.array([
            [-1, 1],
            [ 0, 1],
            [ 1, 1],
        ])

        transformed_points = _transform_points(
            points=points,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )
        # The middle column.
        expected_output = np.array([
            [-1, 0],
            [ 0, 0],
            [ 1, 0],
        ])
        assert np.array_equal(transformed_points, expected_output)

    def test_rotational_position_field(self):

        # Note: applying an affine indicating a clockwise-rotation to a position_field produces a position _ield rotated counter-clockwise.
        # The corresponding effect on a deformed image is a counter-clockwise rotation.

        subject = np.array([
            [0,1,0],
            [0,1,0],
            [0,1,1],
        ])
        subject_resolution = 1
        position_field_resolution = subject_resolution
        # Indicates a 90 degree rotation to the right.
        affine = np.array([
            [ 0, 1, 0],
            [-1, 0, 0],
            [ 0, 0, 1],
        ])
        position_field = _lddmm_utilities._multiply_coords_by_affine(affine, 
            _lddmm_utilities._compute_coords(subject.shape, position_field_resolution))
        # The middle column.
        points = np.array([
            [-1, 0],
            [ 0, 0],
            [ 1, 0],
        ])

        transformed_points = _transform_points(
            points=points,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )
        # The middle row.
        expected_output = np.array([
            [0,  1],
            [0,  0],
            [0, -1],
        ])
        assert np.array_equal(transformed_points, expected_output)

"""
Test lddmm_transform_points.
"""

@pytest.mark.parametrize('deform_to', ['template', 'target'])
class Test_lddmm_transform_points:
    
    def test_identity_position_fields(self, deform_to):
        
        template_shape = (3,4)
        template_resolution = 1
        target_shape = (2,5)
        target_resolution = 1

        affine_phi = _lddmm_utilities._compute_coords(template_shape, template_resolution)
        phi_inv_affine_inv = _lddmm_utilities._compute_coords(target_shape, target_resolution)
        
        if deform_to == 'template':
            points = _lddmm_utilities._compute_coords(target_shape, target_resolution)
            position_field = phi_inv_affine_inv
            position_field_resolution = target_resolution
        else:
            points = _lddmm_utilities._compute_coords(template_shape, template_resolution)
            position_field = affine_phi
            position_field_resolution = template_resolution

        expected_output = _transform_points(
            points=points,
            position_field=position_field,
            position_field_resolution=position_field_resolution,
        )

        transformed_points = lddmm_transform_points(
            points=points,
            deform_to=deform_to,
            affine_phi=affine_phi,
            phi_inv_affine_inv=phi_inv_affine_inv,
            template_resolution=template_resolution,
            target_resolution=target_resolution,
        )

        assert np.array_equal(transformed_points, expected_output)


# if __name__ == "__main__":
#     test_generate_position_field(deform_to='template')
#     test_generate_position_field(deform_to='target')
#     test__transform_image()
#     test_lddmm_transform_image(deform_to='template')
#     test_lddmm_transform_image(deform_to='target')
#     test_lddmm_register()
#     test__transform_points()
#     test_lddmm_transform_points(deform_to='template')
#     test_lddmm_transform_points(deform_to='target')



import numpy as np

class ImageMetaData():
    """Container class for metadata about an image. Stores:
        - nxyz (np.ndarray): image shape - computed from provided image.
        - dxyz (np.ndarray or scalar): image resolution - if dxyz is a scalar, it is upcast to the length of nxyz.
        - xyz (np.ndarray): image coordinates - computed, not provided.
        - origin (np.ndarray): origin for xyz coordinates - default is center.
            Options:
            - 'center'
            - 'zero'
            - np.ndarray of the same length as nxyz
            - scalar, upcast to np.ndarray of same length as nxyz"""
    
    def __init__(self, dxyz, nxyz=None, image=None, origin="center", name=None):
        """If nxyz is provided, use it as nxyz, the image's shape.
        If image is provided, use its shape as nxyz.
        If both nxyz and image are provided and indicate different shapes, raise error.
        If neither nxyz nor image are provided, raise error.
        If origin is provided, it is used to compute xyz.
        If origin is not provided, xyz is centered by default."""

        # Instantiate attributes.
        self.dxyz = None
        self.nxyz = None
        self.xyz = None
        self.name = None

        # Populate attributes.

        # nxyz attribute.
        self.nxyz = ImageMetaData._validate_nxyz(nxyz, image)

        # dxyz attribute.
        self.dxyz = ImageMetaData._validate_dxyz(dxyz, self.nxyz)

        # xyz attribute.
        self.xyz = ImageMetaData._generate_xyz(dxyz, nxyz, origin)

        # name attribute.
        self.name = name


    @staticmethod
    def _validate_nxyz(nxyz, image) -> np.ndarray:
        """Validate compatibility between nxyz and image as provided, 
        and return an appropriate value for the nxyz attribute."""

        # Validate agreement between nxyz and image.
        if nxyz is not None and image is not None:
            if not all(nxyz == np.array(image).shape):
                raise ValueError(f"nxyz and image were both provided, but nxyz does not match the shape of image."
                    f"\nnxyz: {nxyz}, image.shape: {image.shape}.")

        # If nxyz is provided but image is not.
        if nxyz is not None:
            # Cast as np.ndarray.
            if not isinstance(nxyz, np.ndarray):
                nxyz = np.array(nxyz) # Side effect: breaks alias.
            # If nxyz is multidimensional, raise error.
            if nxyz.ndim > 1:
                raise ValueError(f"nxyz cannot be multidimensional.\nnxyz.ndim: {nxyz.ndim}.")
            # If nxyz is 0-dimensional, upcast to 1-dimensional. A perverse case.
            if nxyz.ndim == 0:
                nxyz = np.array([nxyz])
            # If nxyz is 1-dimensional, set nxyz attribute.
            if nxyz.ndim == 1:
                return nxyz
        # If image is provided but nxyz is not.
        # Will fail if image cannot be cast as a np.ndarray.
        elif image is not None:
            # Cast as np.ndarray.
            if not isinstance(image, np.ndarray):
                image = np.array(image)
            # If image is 0-dimensional, upcast to 1-dimensional. A perverse case.
            if image.ndim == 0:
                image = np.array([image])
            # image is a non-zero-dimensional np.ndarray. Set nxyz attribute.
            return np.array(image.shape)
        else:
            raise RuntimeError(f"At least one of nxyz and image must be provided. Both were received as their default value: None.")
        
    
    @staticmethod
    def _validate_dxyz(dxyz, nxyz:np.ndarray) -> np.ndarray:
        """Validate dxyz and its compatibility with nxyz. 
        Return an appropriate value for the dxyz attribute."""

        # Cast as np.ndarray.
        if not isinstance(dxyz, np.ndarray):
            dxyz = np.array(dxyz) # Side effect: breaks alias.
        # if dxyz is multidimensional, raise error.
        if dxyz.ndim > 1:
            raise ValueError(f"dyxz must be 1-dimensional. The value provided is {dxyz.ndim}-dimensional.")
        # If dxyz is 0-dimensional, upcast to match the length of nxyz.
        if dxyz.ndim == 0:
            dxyz = np.array([dxyz]*len(nxyz))
        # dxyz is 1-dimensional.
        if len(dxyz) == len(nxyz):
            # dxyz is 1-dimensional and matches the length of nxyz. Set dxyz attribute.
            return dxyz
        else:
            raise ValueError(f"dyxz must be either 0-dimensional or 1-dimensional and match the length of nxyz or the shape of image."
                f"\nlen(dxyz): {len(dxyz)}.")


    @staticmethod
    def _generate_xyz(dxyz:np.ndarray, nxyz:np.ndarray, origin) -> np.ndarray:
        """Generate image coordinates xyz based on the resolution <dxyz>, the shape <nxyz>, and the center designation <origin>."""

        # Instantiate xyz.
        # xyz is a list of np.ndarray objects of type float and represents the physical coordinates in each dimension.
        xyz = [np.arange(nxyz_i).astype(float)*dxyz_i for nxyz_i, dxyz_i in zip(nxyz, dxyz)]
        
        # Shift xyz.
        if isinstance(origin, str):
            if origin == "center":
                # Offset by the mean along each dimension.
                for coords in xyz:
                    coords -= np.mean(coords)
            elif origin == "zero":
                # This designation is how xyz was initially created and thus it requires no shift.
                pass
            else:
                raise ValueError(f"Unsupported value for origin. Supported string values include ['center', 'zero']."
                    f"\norigin: {origin}.")
        elif isinstance(origin, (int, float, list, np.ndarray)):
            if isinstance(origin, (int, float, list)):
                # Cast to np.ndarray.
                origin = np.array(origin)
            # origin is a np.ndarray.
            # If origin has length 1, broadcast to match nxyz.
            if origin.ndim == 0:
                origin = np.array(list(origin) * len(nxyz))
            # If the length of origin matches the length of nxyz, perform offset in each dimension.
            if len(origin) == len(nxyz):
                for dim, coords in enumerate(xyz):
                    coords -= origin[dim]
            else:
                raise ValueError(f"origin must either be a scalar, have length 1, or match the length of nxyz or the shape of image."
                    f"\nlen(origin): {len(origin)}.")
        else:
            raise ValueError(f"must be one of the following types: [str, int, float, list, np.ndarray]."
                f"\ntype(origin): {type(origin)}.")

        # xyz has been created from dxyz and nxyz, and then adjusted appropriately based on origin.
        return xyz


class Image(ImageMetaData):
    """Subclass of ImageMetaData that also stores the image data itself.
    Attributes:
        - image, the image data
        - dxyz, the resolution in each dimension
        - nxyz, the shape of the image
        - xyz, the coordinates of the image
        - name, optional name"""

    def __init__(self, image, dxyz, origin="center", name=None):
        """Validation is handled by ImageMetaData.__init__."""

        super().__init__(dxyz=dxyz, image=image, origin=origin, name=name)

        # Redo validation on image to set image attribute.
        image = np.array(image) # Side effect: breaks alias
        # If image is 0-dimensional, upcast to 1-dimensional. A perverse case.
        if image.ndim == 0:
            image = np.array([image])
        # Set image attribute.
        self.image = image


r"""
             _____    _____    ______   _   _   _______ 
     /\     |  __ \  |  __ \  |  ____| | \ | | |__   __|
    /  \    | |__) | | |  | | | |__    |  \| |    | |   
   / /\ \   |  _  /  | |  | | |  __|   | . ` |    | |   
  / ____ \  | | \ \  | |__| | | |____  | |\  |    | |   
 /_/    \_\ |_|  \_\ |_____/  |______| |_| \_|    |_|   
                                                        

"""

# This file is read by setup.py and executed as a string.

__name__ = "ardent"
__description__ = "A tool for nonlinear image registration."
__url__ = "https://github.com/neurodata/ardent"
__version__ = "1.0.1"
__author__ = "Devin Crowley"
__author_email__ = "devin.g.crowley@gmail.com"
__license__ = "Apache License 2.0"
__min_python_version__ = (3, 6)
__python_requires__ = ">={}.{}".format(*__min_python_version__)
__classifiers__ = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering :: Mathematics",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]



"""
Defines:
    Internal functions:
        _validate_scalar_to_multi(value, size=None, dtype=None)
        _validate_ndarray(array, dtype=None, minimum_ndim=0, required_ndim=None, 
            forbid_object_dtype=True, broadcast_to_shape=None, reshape_to_shape=None, required_shape=None)
        _validate_resolution(resolution, ndim)
        _compute_axes(shape, resolution=1, origin='center')
        _compute_coords(shape, resolution=1, origin='center')
        _multiply_coords_by_affine(array, affine)
        _compute_tail_determinant(array)
    User functions:
        resample(image, new_resolution, old_resolution=1, 
            err_to_larger=True, extrapolation_fill_value=None, 
            origin='center', method='linear', image_is_coords=False)
        sinc_resample(array, new_shape)
"""

import numpy as np
import warnings

from scipy.interpolate import interpn
from scipy.ndimage import gaussian_filter


def _validate_scalar_to_multi(value, size=None, dtype=None):
    """
    If value's length is 1, upcast it to match size. 
    Otherwise, if it does not match size, raise error.

    If size is not provided, cast to a 1-dimensional np.ndarray.

    Return a numpy array.
    """

    # Cast size to int if appropriate.
    if size is not None:
        try:
            size = int(size)
        except (TypeError, ValueError) as exception:
            raise TypeError(
                f"size must be either None or interpretable as an integer.\n" f"type(size): {type(size)}."
            ) from exception

        if size < 0:
            raise ValueError(f"size must be non-negative.\n" f"size: {size}.")

    # Cast value to np.ndarray.
    try:
        value = np.array(value, dtype=dtype)
    except ValueError as exception:
        raise ValueError(f"value and dtype are incompatible with one another.") from exception

    # Validate value's dimensionality and length.
    if value.ndim == 0:
        value = np.array([value.item()])
    if value.ndim == 1:
        if size is not None and len(value) == 1:
            # Upcast scalar to match size.
            value = np.full(size, value, dtype=dtype)
        elif size is not None and len(value) != size:
            # value's length is incompatible with size.
            raise ValueError(
                f"The length of value must either be 1 or it must match size if size is provided.\n"
                f"len(value): {len(value)}, size: {size}."
            )
    else:
        # value.ndim > 1.
        raise ValueError(
            f"value must not have more than 1 dimension.\n" f"value.ndim: {value.ndim}."
        )

    # Check for np.nan values if the dtype is not object.
    if value.dtype != 'object' and np.any(np.isnan(value)):
        raise NotImplementedError(
            "np.nan values encountered for a value not cast to dtype object. What input led to this result?\n"
            "Write in an exception as appropriate."
        )
        raise ValueError(
            f"value contains inappropriate values for the chosen dtype "
            f"and thus contains np.nan values."
        )

    return value


def _validate_ndarray(
    array,
    dtype=None,
    forbid_object_dtype=True,
    minimum_ndim=1,
    required_ndim=None,
    broadcast_to_shape=None,
    reshape_to_shape=None,
    required_shape=None,
):
    """
    Cast (a copy of) array to a np.ndarray if possible and return it 
    unless it is noncompliant with minimum_ndim, required_ndim, and dtype.
    
    Note: the following checks and validations are performed in order.
    
    If required_ndim is None or 0, _validate_ndarray will accept any object.
    
    array is cast to an np.ndarray of type dtype.

    If minimum_ndim is provided and array.ndim < minimum_ndim, array.shape is left-padded by 1's until minimum_ndim is satisfied.

    If required_ndim is provided and array.ndim != required_ndim, an exception is raised.

    If forbid_object_dtype == True and array.dtype == object, an exception is raised, unless dtype is provided as object.
    
    If a shape is provided to broadcast_to_shape, array is broadcasted to that shape.
    
    If a shape is provided to reshape_to_shape, array is reshaped to that shape.

    If a shape is provided to required_shape, if the shape does not match this shape then an exception is raised.
    """

    # Verify arguments.

    # Verify minimum_ndim.
    if not isinstance(minimum_ndim, int):
        raise TypeError(
            f"minimum_ndim must be of type int.\n"
            f"type(minimum_ndim): {type(minimum_ndim)}."
        )
    if minimum_ndim < 0:
        raise ValueError(
            f"minimum_ndim must be non-negative.\n" f"minimum_ndim: {minimum_ndim}."
        )

    # Verify required_ndim.
    if required_ndim is not None:
        if not isinstance(required_ndim, int):
            raise TypeError(
                f"required_ndim must be either None or of type int.\n"
                f"type(required_ndim): {type(required_ndim)}."
            )
        if required_ndim < 0:
            raise ValueError(
                f"required_ndim must be non-negative.\n"
                f"required_ndim: {required_ndim}."
            )

    # Verify dtype.
    if dtype is not None:
        if not isinstance(dtype, type):
            raise TypeError(
                f"dtype must be either None or a valid type.\n"
                f"type(dtype): {type(dtype)}."
            )

    # Validate array.

    # Cast array to np.ndarray.
    # Validate compliance with dtype.
    try:
        array = np.array(array, dtype) # Side effect: breaks alias.
    except TypeError as exception:
        raise TypeError(
            f"array is of a type that is incompatible with dtype.\n"
            f"type(array): {type(array)}, dtype: {dtype}."
        ) from exception
    except ValueError as exception:
        raise ValueError(
            f"array has a value that is incompatible with dtype.\n"
            f"array: {array}, \ntype(array): {type(array)}, dtype: {dtype}."
        ) from exception

    # Verify compliance with forbid_object_dtype.
    if forbid_object_dtype:
        if array.dtype == object and dtype != object:
            raise TypeError(
                f"Casting array to a np.ndarray produces an array of dtype object \n"
                f"while forbid_object_dtype == True and dtype != object."
            )

    # Validate compliance with minimum_ndim by left-padding the shape with 1's as necessary.
    if array.ndim < minimum_ndim:
        array = array.reshape(*[1] * (minimum_ndim - array.ndim), *array.shape)

    # Validate compliance with required_ndim.
    if required_ndim is not None and array.ndim != required_ndim:
        raise ValueError(
            f"If required_ndim is not None, array.ndim must be made to equal it.\n"
            f"array.ndim: {array.ndim}, required_ndim: {required_ndim}."
        )

    # Broadcast array if appropriate.
    if broadcast_to_shape is not None:
        array = np.copy(np.broadcast_to(array=array, shape=broadcast_to_shape))

    # Reshape array if appropriate.
    if reshape_to_shape is not None:
        array = np.copy(array.reshape(reshape_to_shape))

    # Verify compliance with required_shape if appropriate.
    if required_shape is not None:
        try:
            required_shape_satisfied = np.array_equal(array.reshape(required_shape).shape, array.shape)
        except ValueError as exception:
            raise ValueError(
                f"array is incompatible with required_shape.\n"
                f"array.shape: {array.shape}, required_shape: {required_shape}."
            ) from exception
        if not required_shape_satisfied:
            raise ValueError(f"array is compatible with required_shape but does not match required_shape.\n"
                             f"array.shape: {array.shape}, required_shape: {required_shape}.")

    return array


def _validate_resolution(resolution, ndim, dtype=float):
    """Validate resolution to assure its length matches the dimensionality of image."""

    resolution = _validate_scalar_to_multi(resolution, size=ndim, dtype=dtype)

    if np.any(resolution <= 0):
        raise ValueError(
            f"All elements of resolution must be positive.\n"
            f"np.min(resolution): {np.min(resolution)}."
        )

    return resolution


def _compute_axes(shape, resolution=1, origin="center"):
    """
    Returns the real_axes defining an image with the given shape 
    at the given resolution as a list of numpy arrays.
    """

    # Validate shape.
    shape = _validate_ndarray(shape, dtype=int, required_ndim=1)

    # Validate resolution.
    resolution = _validate_resolution(resolution, len(shape))

    # Create axes.

    # axes is a list of arrays matching each shape element from shape, spaced by the corresponding resolution.
    axes = [
        np.arange(dim_size) * dim_res for dim_size, dim_res in zip(shape, resolution)
    ]

    # List all presently recognized origin values.
    origins = ["center", "zero"]

    if origin == "center":
        # Center each axes array to its mean.
        for axis_index, axis in enumerate(axes):
            axes[axis_index] -= np.mean(axis)
    elif origin == "zero":
        # Allow each axis to increase from 0 along each dimension.
        pass
    else:
        raise NotImplementedError(
            f"origin must be one of these supported values: {origins}.\n"
            f"origin: {origin}."
        )

    return axes


def _compute_coords(shape, resolution=1, origin="center"):
    """
    Returns the real_coordinates of an image with the given shape 
    at the given resolution as a single numpy array of shape (*shape, len(shape)).
    """

    axes = _compute_axes(shape, resolution, origin)

    meshes = np.meshgrid(*axes, indexing="ij")

    return np.stack(meshes, axis=-1)


def _multiply_coords_by_affine(affine, array):
    """Applies affine to the elements of array at each spatial position and returns the result."""

    # Validate inputs.

    # Verify that affine is square.
    if affine.ndim != 2:
        raise ValueError(
            f"affine must be a 2-dimensional matrix.\n" f"affine.ndim: {affine.ndim}."
        )
    # affine is 2-dimensional.
    if affine.shape[0] != affine.shape[1]:
        raise ValueError(
            f"affine must be a square matrix.\n" f"affine.shape: {affine.shape}."
        )
    # affine is square.

    # Verify compatibility between affine and array.
    if array.shape[-1] != len(affine) - 1:
        raise ValueError(
            f"array is incompatible with affine. The length of the last dimension of array should be 1 less than the length of affine.\n"
            f"array.shape: {array.shape}, affine.shape: {affine.shape}."
        )

    # Raise warning if affine is not in homogenous coordinates.
    if not np.array_equal(affine[-1], np.array([0] * (len(affine) - 1) + [1])):
        warnings.warn(
            message=f"affine is not in homogenous coordinates.\n"
                    f"affine[-1] should be zeros with a 1 on the right.\n"
                    f"affine[-1]: {affine[-1]}.",
            category=RuntimeWarning,
        )

    # Perform affine matrix multiplication.

    ndims = len(affine) - 1
    return (
        np.squeeze(np.matmul(affine[:ndims, :ndims], array[..., None]), -1)
        + affine[:ndims, ndims]
    )


def _compute_tail_determinant(array):
    """Computes and returns the determinant of array along its last 2 dimensions."""

    # Validate that array is square on its last 2 dimensions.
    if array.shape[-1] != array.shape[-2]:
        raise ValueError(f"array must be square on its last 2 dimensions.\n"
                         f"array.shape[-2:]: {array.shape[-2:]}.")

    # Compute the determinant recursively.

    if array.shape[-1] == 2:
        # Handle 2-dimensional base case.
        determinant = array[...,0,0] * array[...,1,1] - array[...,0,1] * array[...,1,0]
    else:
        # Handle more than 2-dimensional recursive case.
        determinant = 0
        for dim in range(array.shape[-1]):
            recursive_indices = list(range(array.shape[-1]))
            recursive_indices.remove(dim)
            determinant += (-1)**dim * array[...,0,dim] * _compute_tail_determinant(array[...,1:,recursive_indices])
    
    return determinant


def resample(
    image,
    new_resolution,
    old_resolution=1,
    err_to_larger=True,
    extrapolation_fill_value=None,
    origin="center",
    method="linear",
    image_is_coords=False,
    anti_aliasing=True,
):
    """
    Resamples image from an old resolution to a new resolution.
    
    Parameters
    ----------
        image: np.ndarray
            The image to be resampled
        new_resolution: float, seq
            The resolution of the resampled image.
        old_resolution: float, seq, optional
            The resolution of the input image. By default 1.
        err_to_larger: bool, optional
            Determines whether to round the new shape up or down. By default True.
        extrapolation_fill_value: float, optional
            The fill_value kwarg passed to interpn. By default None.
        origin: str, optional
            The origin to use for the image axes and coordinates used internally. By default 'center'.
        method: str, optional
            The method of interpolation, passed as the method kwarg in interpn. By default 'linear'.
        image_is_coords: bool, optional
            If True, this implies that the last dimension of image is not a spatial dimension and not subject to interpolation. By default False.
        anti_aliasing: bool, optional
            If True, applies a gaussian filter across dimensions to be downsampled before interpolating. By default True.
    
    Returns
    -------
    np.ndarray
        The result of resampling image at new_resolution.
    """

    # Validate inputs and define ndim & old_shape based on image_is_coords.
    image = _validate_ndarray(image) # Breaks alias.
    if image_is_coords:
        ndim = image.ndim - 1
        old_shape = image.shape[:-1]
    else:
        ndim = image.ndim
        old_shape = image.shape
    new_resolution = _validate_resolution(new_resolution, ndim)
    old_resolution = _validate_resolution(old_resolution, ndim)

    # Handle trivial case.
    if np.array_equal(new_resolution, old_resolution):
        return (
            image # Note: this is a copy of the input image and is not the same object.
        )

    # Compute new_coords and old_axes.
    if err_to_larger:
        new_shape = np.ceil(old_shape * old_resolution / new_resolution)
    else:
        new_shape = np.floor(old_shape * old_resolution / new_resolution)
    new_coords = _compute_coords(new_shape, new_resolution, origin)
    old_axes = _compute_axes(old_shape, old_resolution, origin)

    # Apply anti-aliasing gaussian filter if downsampling.
    if anti_aliasing:
        if image_is_coords:
            downsample_factors = np.insert(
                old_shape / new_shape, image.ndim - 1, values=0, axis=0
            )
        else:
            downsample_factors = old_shape / new_shape
        anti_aliasing_sigma = np.maximum(0, (downsample_factors - 1) / 2)
        gaussian_filter(
            image, anti_aliasing_sigma, output=image, mode="nearest"
        )  # Mutates image.

    # Interpolate image.
    new_image = interpn(
        points=old_axes,
        values=image,
        xi=new_coords,
        bounds_error=False,
        fill_value=extrapolation_fill_value,
    )

    return new_image


def sinc_resample(array, new_shape):
    """
    Resample array to new_shape by padding and truncating at high frequencies in the fourier domain.

    Parameters
    ----------
    array : np.ndarray
        The array to be resampled.
    new_shape : seq
        The shape to resample array to.

    Returns
    -------
    np.ndarray
        A copy of array after resampling.
    """

    # Validate inputs.
    array = _validate_ndarray(array)
    new_shape = _validate_ndarray(new_shape, dtype=int, required_ndim=1, required_shape=array.ndim)

    resampled_array = np.copy(array)

    for dim in range(array.ndim):
        fourier_transformed_array = np.fft.rfft(resampled_array, axis=dim)
        resampled_array = np.fft.irfft(fourier_transformed_array, axis=dim, n=new_shape[dim])
    resampled_array *= resampled_array.size / array.size

    return resampled_array



from . import lddmm # Subpackage.
from .lddmm import resample # Function.

from .transform import Transform # Class.

from .visualization import heatslices # Function.

from .file_io import save, load # Functions.

from . import preprocessing # Subpackage.
from .preprocessing import preprocess # Function.

from . import presets # Subpackage.
from .presets import basic_preprocessing # Function.
from .presets import basic_preprocessing_with_pad # Function.
from .presets import get_registration_presets # Function.
from .presets import get_registration_preset # Function.



import numpy as np
from numpy import ma
import SimpleITK as sitk
from scipy.ndimage.filters import gaussian_filter
from skimage import morphology
from skimage.filters import threshold_otsu
from skimage.transform import resize, rescale

from ..lddmm._lddmm_utilities import _validate_ndarray


def correct_bias_field(image, correct_at_scale=1, as_float32=True, **kwargs):
    """
    Shifts image such that its minimum value is 1, computes the bias field after downsampling by correct_at_scale, 
    upsamples this bias field and applies it to the shifted image, then undoes the shift and returns the result.
    Computes bias field using sitk.N4BiasFieldCorrection (http://bit.ly/2oFwAun).
    
    Args:
        image (np.ndarray): The image to be bias corrected.
        correct_at_scale (float, optional): The scale by which the shape of image is reduced before computing the bias. Defaults to 1.
        as_float32 (bool, optional): If True, image is internally cast as a sitk.Image of type sitkFloat32. If False, it is of type sitkFloat64. Defaults to True.

    Kwargs:
        Any additional keyword arguments overwrite the default values passed to sitk.N4BiasFieldCorrection.
    
    Returns:
        np.ndarray: A copy of image after bias correction.
    """

    # Validate inputs.

    # Validate image.
    image = _validate_ndarray(image, dtype=float)

    # Verify correct_at_scale.
    correct_at_scale = float(correct_at_scale)
    if correct_at_scale < 1:
        raise ValueError(f"correct_at_scale must be equal to or greater than 1.\n"
                         f"correct_at_scale: {correct_at_scale}.")

    # Shift image such that its minimum value lies at 1.
    image_min = image.min()
    image = image - image_min + 1

    # Downsample image according to scale.
    downsampled_image = rescale(image, correct_at_scale)

    # Bias correct downsampled_image.
    N4BiasFieldCorrection_kwargs = dict(
        image=downsampled_image, 
        maskImage=np.ones_like(downsampled_image), 
        convergenceThreshold=0.001, 
        maximumNumberOfIterations=[50, 50, 50, 50], 
        biasFieldFullWidthAtHalfMaximum=0.15, 
        wienerFilterNoise=0.01, 
        numberOfHistogramBins=200,
        numberOfControlPoints=[4, 4, 4], 
        splineOrder=3, 
        useMaskLabel=True, 
        maskLabel=1, 
    )
    # Overwrite default arguments with user-supplied kwargs.
    N4BiasFieldCorrection_kwargs.update(kwargs)
    # Convert image and maskImage N4BiasFieldCorrection_kwargs from type np.ndarray to type sitk.Image.
    sitk_image = sitk.GetImageFromArray(N4BiasFieldCorrection_kwargs['image'])
    sitk_image = sitk.Cast(sitk_image, sitk.sitkFloat32 if as_float32 else sitk.sitkFloat64)
    sitk_maskImage = N4BiasFieldCorrection_kwargs['maskImage'].astype(np.uint8)
    sitk_maskImage = sitk.GetImageFromArray(sitk_maskImage)
    N4BiasFieldCorrection_kwargs.update(
        image=sitk_image,
        maskImage=sitk_maskImage,
    )
    bias_corrected_downsampled_image = sitk.N4BiasFieldCorrection(*N4BiasFieldCorrection_kwargs.values())
    bias_corrected_downsampled_image = sitk.GetArrayFromImage(bias_corrected_downsampled_image)

    # Compute bias from bias_corrected_downsampled_image.
    downsample_computed_bias = bias_corrected_downsampled_image / downsampled_image

    # Upsample bias.
    upsampled_bias = resize(downsample_computed_bias, image.shape)

    # Apply upsampled bias to original resolution shifted image.
    bias_corrected_image = image * upsampled_bias

    # Reverse the initial shift.
    bias_corrected_image += image_min - 1

    return bias_corrected_image


def remove_grid_artifact(image, z_axis=0, sigma_blur=None, mask='Otsu', otsu_nbins=256, otsu_binary_closing_radius=None, otsu_background_is_dim=True):
    """
    Remove the grid artifact from tiled data.
    
    Args:
        image (np.ndarray): The image with a grid artifact.
        z_axis (int, optional): The axis along which the tiles are stacked. Defaults to 0.
        sigma_blur (float, optional): The size of the blur used to compute the bias for grid edges in units of voxels. 
            Should be approximately the size of the tiles. Defaults to np.ceil(np.sqrt(np.min(image.shape))).
        mask (np.ndarray, str, NoneType, optional): A mask of the valued voxels in the image. 
            Supported values are:
                a np.ndarray with a shape corresponding to image.shape, 
                None, indicating no mask (i.e. all voxels are considered in the artifact correction), 
                'Otsu', indicating that the Otsu threshold will be used to identify foreground and background voxels.
            Defaults to 'Otsu'.
        otsu_nbins (int, optional): The number of bins used to calculate the histogram in skimage.filters.threshold_otsu if mask == 'Otsu'. Defaults to 256.
        otsu_binary_closing_radius (int, optional): The radius of the structuring element given to binary_close if mask == 'Otsu'. Defaults to np.ceil(np.sqrt(np.min(image.shape)) / 3).
        otsu_background_is_dim (bool, optional): If True and mask == 'Otsu', when computing the mask it is assumed that the background will have a lower value than the foreground. Defaults to True.
    
    Returns:
        np.ndarray: A copy of image with its grid artifact removed.
    """

    # Validate inputs.

    # Validate image.
    image = _validate_ndarray(image, dtype=float)

    # Validate sigma_blur.
    sigma_blur = float(sigma_blur) if sigma_blur is not None else np.ceil(np.sqrt(np.min(image.shape)))

    # Validate otsu_binary_closing_radius.
    otsu_binary_closing_radius = int(otsu_binary_closing_radius) if otsu_binary_closing_radius is not None else int(np.ceil(np.sqrt(np.min(image.shape)) / 3))

    # Validate otsu_background_is_dim.
    otsu_background_is_dim = bool(otsu_background_is_dim)

    # Construct masked_image as a ma.MaskedArray.

    # Interpret input mask.
    if mask is None:
        mask = np.ones_like(image, bool)
    elif isinstance(mask, str) and mask == 'Otsu':
        # Finds the optimal split threshold between the foreground anad background, 
        # by maximizing the interclass variance and minimizing the intraclass variance between voxel intensities, 
        # with he higher-intensity class labeled as 1.
        otsu_threshold = threshold_otsu(image, nbins=otsu_nbins)
        mask = np.ones_like(image, int)
        # Segment image by otsu_threshold.
        mask[image <= otsu_threshold if otsu_background_is_dim else image >= otsu_threshold] = 0
        # Perform binary closing and then binary fill hole on image to remove mislabed background voxels inside the foreground regions.
        mask = sitk.GetArrayFromImage(
            sitk.BinaryFillhole(
                sitk.BinaryMorphologicalClosing(
                    sitk.GetImageFromArray(mask), 
                    otsu_binary_closing_radius, 
                    sitk.sitkBall
                )
            )
        ).astype(bool)
    else:
        mask = _validate_ndarray(mask, reshape_to_shape=image.shape, dtype=bool)
    
    # Use the inverse of mask to create masked_image.
    masked_image = ma.masked_array(image, mask=~mask)

    # Shift masked_image so that its minimum lies at 1.
    masked_image_min = np.min(masked_image)
    masked_image -= masked_image_min - 1

    # Correct grid artifacts.

    # Take the average across z
    mean_across_z = np.mean(masked_image, axis=z_axis, keepdims=True)

    # Blur the mean with a gaussian.
    z_projection_bias = gaussian_filter(mean_across_z, sigma_blur) / mean_across_z

    # Apply the z_projection_bias to correct the masked_image.
    corrected_masked_image = masked_image * z_projection_bias

    # Reverse the shift to restore the original data window.
    corrected_masked_image += masked_image_min - 1

    # Return the unmasked array.
    return corrected_masked_image.data


from .batch_preprocessing import basic_preprocessing
from .batch_preprocessing import basic_preprocessing_with_pad
from .registration_parameters import get_registration_preset
from .registration_parameters import get_registration_presets


import numpy as np

from ..lddmm._lddmm_utilities import _validate_ndarray

def _verify_data_is_ndarray(data):
    if not isinstance(data, np.ndarray):
        raise TypeError(f"data must be of type np.ndarray.\ntype(data): {type(data)}.")


def cast_to_typed_array(data, dtype=float):
    """
    Returns a copy of data cast as a np.ndarray of type dtype.
    
    Args:
        data (np.ndarray): The array to be cast.
        dtype (type, optional): The dtype to cast data to. Defaults to float. Defaults to float.
    
    Returns:
        np.ndarray: A copy of data cast to type dtype.
    """

    return _validate_ndarray(data, dtype=dtype)


def normalize_by_MAD(data):
    """
    Returns a copy of data divided by its mean absolute deviation.
    
    Args:
        data (np.ndarray): The array to be normalized.
    
    Returns:
        np.ndarray: A copy of data divided by its mean absolute deviation.
    """
    
    _verify_data_is_ndarray(data)

    mean_absolute_deviation = np.mean(np.abs(data - np.median(data)))

    normalized_data = data / mean_absolute_deviation

    return normalized_data


def center_to_mean(data):
    """
    Returns a copy of data subtracted by its mean.

    Args:
        data (np.ndarray): The array to be subtracted from.
    Returns:
        np.ndarray: A copy of data subtracted by its mean.
    """

    _verify_data_is_ndarray(data)

    centered_data = data - np.mean(data)

    return centered_data


def pad(data, pad_width=10, mode='constant', constant_values=None):
    """
    Returns a padded copy of data.
    
    Args:
        data (np.ndarray): The array to be padded.
        pad_width (int, optional): The amount by which to pad. Defaults to 10.
        mode (str, optional): The padding mode used in np.pad. Defaults to 'constant'.
        constant_values (float, optional): The values to use in padding if mode='constant' If None, this is set to np.quantile(data, 10**-data.ndim). Defaults to None.
    
    Returns:
        np.ndarray: The padded copy of data.
    """

    _verify_data_is_ndarray(data)

    if constant_values is None:
        constant_values = np.quantile(data, 10**-data.ndim)

    pad_kwargs = {'array':data, 'pad_width':pad_width, 'mode':mode}
    if mode == 'constant': pad_kwargs.update(constant_values=constant_values)

    padded_data = np.pad(**pad_kwargs)

    return padded_data





import numpy as np
import SimpleITK as sitk
from pathlib import Path
import pickle

def _validate_inputs(**kwargs):
    """Accepts arbitrary kwargs. If recognized, they are validated.
    If they cannot be validated an exception is raised.
    A dictionary containing the validated kwargs is returned."""

    # Validate data.

    if 'data' in kwargs.keys():
        data = kwargs['data']
        if isinstance(data, dict):
            # Verify that each value in the dict is a np.ndarray object and recursion will stop at 1 level.
            if not all(map(lambda datum: isinstance(datum, np.ndarray), data.values())):
                raise TypeError(f"data must be either a np.ndarray or a dictionary mapping only to np.ndarray objects.")
            # Recurse into each element of the dictionary keys of np.ndarray objects.
            for key, datum in data.items():
                data[key] = _validate_inputs(data=datum)['data']
        # If data is neither a list, dict, nor a np.ndarray, raise TypeError.
        elif not isinstance(data, np.ndarray):
            raise TypeError(f"data must be a np.ndarray.\ntype(data): {type(data)}.")
        # data is a np.ndarray.
        kwargs.update(data=data)

    # Validate file_path.

    if 'file_path' in kwargs.keys():
        file_path = kwargs['file_path']
        file_path = Path(file_path).expanduser().resolve() # Will raise exceptions if file_path is not an oppropriate type.
        if not file_path.parent.is_dir():
            raise FileNotFoundError(f"file_path corresponds to a location that does not presently exist.\n"
                f"file_path.parent: {file_path.parent}.")
        # Validate extension.
        if not file_path.suffix:
            default_extension = '.vtk'
            file_path = file_path.with_suffix(default_extension)
        # file_path is valid.
        kwargs.update(file_path=file_path)

    return kwargs


def save(data, file_path):
    """
    Save data to file_path
    
    Args:
        data (np.ndarray, dict): An array or a dictionary with np.ndarray values to be saved.
        file_path (str, Path): The file path that data will be saved to. Accepts an arbitrary suffix but provides .vtk by default.
    
    Raises:
        Exception: Raised if _validate_inputs has failed to catch an improper type for data.
    """
    
    # Validate inputs.
    inputs = {'data':data, 'file_path':file_path}
    validated_inputs = _validate_inputs(**inputs)
    data = validated_inputs['data']
    file_path = validated_inputs['file_path']

    # data is either a single np.ndarray or a list of np.ndarrays.

    if isinstance(data, np.ndarray):
        # Convert data to sitk.Image.
        # Take the transpose of data to premptively correct for the f-ordering of SimpleITK Images.
        data = data.T
        data_Image = sitk.GetImageFromArray(data) # Side effect: breaks alias.
        # Save data to file_path.
        sitk.WriteImage(data_Image, str(file_path))
    # If data is a dictionary it must map to np.ndarray objects.
    elif isinstance(data, dict):
        np.savez(file_path.with_suffix(''), **data) # '.npz' is appended.
    else:
        # _validate_inputs has failed.
        raise Exception(f"_validate_inputs has failed to prevent an improper type for data.\n"
            f"type(data): {type(data)}.")


def load(file_path):
    """
    Load data from file_path.
    
    Args:
        file_path (str, Path): The file path from which data will be retrieved.
    
    Returns:
        np.ndarray, dict: The array or dict of arrays saved at file_path.
    """

    # Validate inputs.
    inputs = {'file_path':file_path}
    validated_inputs = _validate_inputs(**inputs)
    file_path = validated_inputs['file_path']

    if file_path.suffix == '.npz':
        data = np.load(file_path)
        # data is a dictionary.
        return data
    else:
        # Read in data as sitk.Image.
        data_Image = sitk.ReadImage(str(file_path))
        # Convert data_Image to np.ndarray.
        data = sitk.GetArrayFromImage(data_Image)
        # Take the transpose of data to correct for the f-ordering of SimpleITK Images.
        data = data.T
        # data is a np.ndarray.
        return data


def save_pickled(obj, file_path):
    """
    Pickle object obj and save it to file_path.
    
    Args:
        obj (object): The pickleable object to be saved.
        file_path (str, Path): The file path at which to save obj.
    """

    # Validate file_path.
    file_path = _validate_inputs(file_path=file_path)['file_path']

    with open(file_path, 'wb') as file:
        pickle.dump(obj, file)


def load_pickled(file_path):
    """
    Load pickled object from file_path.
    
    Args:
        file_path (str, Path): The file path at which a pickled object is saved.
    
    Returns:
        object: The pickled object saved at file_path.
    """

    # Validate file_path.
    file_path = _validate_inputs(file_path=file_path)['file_path']

    # TODO: verify behavior of suffixes.
    with open(file_path, 'rb') as file:
        return pickle.load(file)


from ..preprocessing import preprocess

def basic_preprocessing(data):
    """
    Call each of the listed preprocessing functions on <data> and return the result.
    Preprocessing functions:

        - cast_to_typed_array
        - normalize_by_MAD
        - center_to_mean

    Args:
        data (np.ndarray, list): An array to be preprocessed, or a list of arrays to be preprocessed.
    
    Returns:
        np.ndarray, list: The input array <data> after preprocessing, or each preprocessed array from the input list <data> if a list of arrays was provided.
    """

    return preprocess(data, ['cast_to_typed_array', 'normalize_by_MAD', 'center_to_mean'])


def basic_preprocessing_with_pad(data):
    """
    Call each of the listed preprocessing functions on <data> and return the result.
    Preprocessing functions:

        - cast_to_typed_array
        - pad
        - normalize_by_MAD
        - center_to_mean

    Args:
        data (np.ndarray, list): An array to be preprocessed, or a list of arrays to be preprocessed.
    
    Returns:
        np.ndarray, list: The input array <data> after preprocessing, or each preprocessed array from the input list <data> if a list of arrays was provided.
    """

    return preprocess(data, ['cast_to_typed_array', 'pad', 'normalize_by_MAD', 'center_to_mean'])



# Transform is the primary class in ARDENT.

import numpy as np
from pathlib import Path
import pickle

from .presets import get_registration_preset
from .lddmm._lddmm import lddmm_register, lddmm_transform_image, lddmm_transform_points
from . import file_io


class Transform:
    """
    Transform stores the deformation that is output by a registration 
    and provides methods for applying that transformation to various images.
    """
    
    def __init__(self):
        """
        Initialize Transform object. Sets attributes to None.
        """

        # Create attributes.

        # Saved for the continue_registration method.
        self._registration_parameters = None

        # lddmm_dict.

        # Core.
        self.affine=None,
        self.phi=None,
        self.phi_inv=None,
        self.affine_phi=None,
        self.phi_inv_affine_inv=None,
        self.contrast_coefficients=None,
        self.velocity_fields=None

        # Helpers.
        self.template_resolution=None,
        self.target_resolution=None,
        
        # Accumulators.
        self.matching_energies=None,
        self.regularization_energies=None,
        self.total_energies=None,

        # Debuggers.
        self.lddmm=None,
        

    def _update_lddmm_attributes(self, lddmm_dict):
        """Update attributes with the output dictionary from lddmm_register."""

        # Verify lddmm_dict.
        if not isinstance(lddmm_dict, dict):
            raise TypeError(f"lddmm_dict must be of type dict.\n"
                            f"type(lddmm_dict): {type(lddmm_dict)}.")

        # Set attributes.

        for key, value in lddmm_dict.items():
            if not hasattr(self, key):
                raise ValueError(f"lddmm_dict must only have keys that are attributes of self.\n"
                                 f"key: {key}.")
            setattr(self, key, value)


    def get_lddmm_dict(self):
        """
        Constructs lddmm_dict, a dictionary of this object's attributes as populated by the return from lddmm_register.
        
        Returns:
            dict: The attributes of this Transform object, matching the dictionary returned from lddmm_register.
        """

        return dict(
            # Core.
            affine=self.affine,
            phi=self.phi,
            phi_inv=self.phi_inv,
            affine_phi=self.affine_phi,
            phi_inv_affine_inv=self.phi_inv_affine_inv,
            contrast_coefficients=self.contrast_coefficients,
            velocity_fields=self.velocity_fields,
            # Helpers.
            template_resolution=self.template_resolution,
            target_resolution=self.target_resolution,
            # Accumulators.
            matching_energies=self.matching_energies,
            regularization_energies=self.regularization_energies,
            total_energies=self.total_energies,
            # Debuggers.
            lddmm=self.lddmm,
        )


    def register(
        self,
        # Images.
        template,
        target,
        # Image resolutions.
        template_resolution=None,
        target_resolution=None,
        # Multiscale.
        multiscales=None,
        # Preset.
        preset=None,
        # Iterations.
        num_iterations=None,
        num_affine_only_iterations=None,
        num_rigid_affine_iterations=None,
        # Stepsizes.
        affine_stepsize=None,
        deformative_stepsize=None,
        # Affine specifiers.
        fixed_affine_scale=None,
        # Velocity field specifiers.
        sigma_regularization=None,
        velocity_smooth_length=None,
        preconditioner_velocity_smooth_length=None,
        maximum_velocity_fields_update=None,
        num_timesteps=None,
        # Contrast map specifiers.
        contrast_order=None,
        spatially_varying_contrast_map=None,
        contrast_iterations=None,
        sigma_contrast=None,
        contrast_smooth_length=None,
        # Smoothness vs. accuracy tradeoff.
        sigma_matching=None,
        # Classification specifiers.
        classify_and_weight_voxels=None,
        sigma_artifact=None,
        sigma_background=None,
        artifact_prior=None,
        background_prior=None,
        # Initial values.
        initial_affine=None,
        initial_velocity_fields=None,
        initial_contrast_coefficients=None,
        # Diagnostic outputs.
        calibrate=None,
        track_progress_every_n=None,
    ):
        """
        Compute a registration between template and target, to be applied with lddmm_transform_image.
        
        Args:
            template (np.ndarray): The ideally clean template image being registered to the target.
            target (np.ndarray): The potentially messier target image being registered to.
            template_resolution (float, list, optional): A scalar or list of scalars indicating the resolution of the template. Overrides 0 input. Defaults to 1.
            target_resolution (float, optional): A scalar or list of scalars indicating the resolution of the target. Overrides 0 input. Defaults to 1.
            multiscales (float, seq, optional): A scalar, list of scalars, or list of lists or np.ndarray of scalars, determining the levels of downsampling at which the registration should be performed before moving on to the next. 
                Values must be either all at least 1, or all at most 1. Both options are interpreted as downsampling. For example, multiscales=[10, 3, 1] will result in the template and target being downsampled by a factor of 10 and registered. 
                This registration will be upsampled and used to initialize another registration of the template and target downsampled by 3, and then again on the undownsampled data. multiscales=[1/10, 1/3, 1] is equivalent. 
                Alternatively, the scale for each dimension can be specified, e.g. multiscales=[ [10, 5, 5], [3, 3, 3], 1] for a 3D registration will result in the template and target downsampled by [10, 5, 5], then [3, 3, 3], then [1, 1, 1]. 
                If provided with more than 1 value, all following arguments with the exceptions of initial_affine, initial_velocity_fields, and initial_contrast_coefficients, 
                which may be provided for the first value in multiscales, may optionally be provided as sequences with length equal to the number of values provided to multiscales. Each such value is used at the corresponding scale. 
                Additionally, template_resolution and target_resolution cannot be provided for each scale in multiscales. Rather, they are given once to indicate the resolution of the template and target as input.
                multiscales should be provided as descending values. Defaults to 1.
            preset (str, optional): A string specifying a recognized preset, or subset of the following arguments to provide automatically, overridden by values specified in this call. Defaults to None.
            num_iterations (int, optional): The total number of iterations. Defaults to 300.
            num_affine_only_iterations (int, optional): The number of iterations at the start of the process without deformative adjustments. Defaults to 100.
            num_rigid_affine_iterations (int, optional): The number of iterations at the start of the process in which the affine is kept rigid. Defaults to 50.
            affine_stepsize (float, optional): The unitless stepsize for affine adjustments. Should be between 0 and 1. Defaults to 0.3.
            deformative_stepsize (float, optional): The stepsize for deformative adjustments. Optimal values are problem-specific. Setting preconditioner_velocity_smooth_length increases the appropriate value of deformative_stepsize. 
                If equal to 0 then the result is affine-only registration. By default 0.
            fixed_affine_scale (float, optional): The scale to impose on the affine at all iterations. If None, no scale is imposed. Otherwise, this has the effect of making the affine always rigid. Defaults to None.
            sigma_regularization (float, optional): A scalar indicating the freedom to deform. Small values put harsher constraints on the smoothness of a deformation. 
                With sufficiently large values, the registration will overfit any noise in the target, leading to unrealistic deformations. 
                However, this may still be appropriate with a small num_iterations. Overrides 0 input. Defaults to np.inf.
            velocity_smooth_length (float, optional): The length scale of smoothing. Overrides 0 input. Defaults to 2 * np.max(self.template_resolution).
            preconditioner_velocity_smooth_length (float, optional): The length of preconditioner smoothing of the velocity_fields in physical units. Affects the optimization of the velocity_fields, but not the optimum. Defaults to 0.
            maximum_velocity_fields_update (float, optional): The maximum allowed update to the velocity_fields in physical units. Affects the optimization of the velocity_fields, but not the optimum. 
                Overrides 0 input. Defaults to np.max(self.template.shape * self.template_resolution).
            num_timesteps (int, optional): The number of composed sub-transformations in the diffeomorphism. Overrides 0 input. Defaults to 5.
            contrast_order (int, optional): The order of the polynomial fit between the contrasts of the template and target. Overrides 0 input. Defaults to 1.
            spatially_varying_contrast_map (bool, optional): If True, uses a polynomial per voxel to compute the contrast map rather than a single polynomial. Defaults to False.
            contrast_iterations (int, optional): The number of iterations of gradient descent to converge toward the optimal contrast_coefficients if spatially_varying_contrast_map == True. Overrides 0 input. Defaults to 5.
            sigma_contrast (float, optional): The scale of variation in the contrast_coefficients if spatially_varying_contrast_map == True. Overrides 0 input. Defaults to 1e-2.
            contrast_smooth_length (float, optional): The length scale of smoothing of the contrast_coefficients if spatially_varying_contrast_map == True. Overrides 0 input. Defaults to 2 * np.max(self.target_resolution).
            sigma_matching (float, optional): An estimate of the spread of the noise in the target, 
                representing the tradeoff between the regularity and accuracy of the registration, where a smaller value should result in a less smooth, more accurate result. 
                Typically it should be set to an estimate of the standard deviation of the noise in the image, particularly with artifacts. Overrides 0 input. Defaults to the standard deviation of the target.
            classify_and_weight_voxels (bool, optional): If True, artifacts are jointly classified with registration using sigma_artifact. Defaults to False.
            sigma_artifact (float, optional): The level of expected variation between artifact and non-artifact intensities. Overrides 0 input. Defaults to 5 * sigma_matching.
            sigma_background (float, optional): The level of expected variation between background and non-background intensities. Overrides 0 input. Defaults to 2 * sigma_matching.
            artifact_prior (float, optional): The prior probability at which we expect to find that any given voxel is artifact. Defaults to 1/3.
            background_prior (float, optional): The prior probability at which we expect to find that any given voxel is background. Defaults to 1/3.
            initial_affine (np.ndarray, optional): The affine array that the registration will begin with. Defaults to np.eye(template.ndim + 1).
            initial_velocity_fields (np.ndarray, optional): The velocity fields that the registration will begin with. Defaults to all zeros.
            initial_contrast_coefficients (np.ndarray, optional): The contrast coefficients that the registration will begin with. 
                If None, the 0th order coefficient(s) are set to np.mean(self.target) - np.mean(self.template) * np.std(self.target) / np.std(self.template), 
                if self.contrast_order > 1, the 1st order coefficient(s) are set to np.std(self.target) / np.std(self.template), 
                and all others are set to zero. Defaults to None.
            calibrate (bool, optional): A boolean flag indicating whether to accumulate additional intermediate values and display informative plots for calibration purposes. Defaults to False.
            track_progress_every_n (int, optional): If positive, a progress update will be printed every track_progress_every_n iterations of registration. Defaults to 0.
        """

        # Collect registration parameters.
        registration_parameters = dict(
            # Images.
            template=template,
            target=target,
            # Image resolutions.
            template_resolution=template_resolution,
            target_resolution=target_resolution,
            # Multiscale.
            multiscales=multiscales,
            # Iterations.
            num_iterations=num_iterations,
            num_affine_only_iterations=num_affine_only_iterations,
            num_rigid_affine_iterations=num_rigid_affine_iterations,
            # Stepsizes.
            affine_stepsize=affine_stepsize,
            deformative_stepsize=deformative_stepsize,
            # Affine specifiers.
            fixed_affine_scale=fixed_affine_scale,
            # Velocity field specifiers.
            sigma_regularization=sigma_regularization,
            velocity_smooth_length=velocity_smooth_length,
            preconditioner_velocity_smooth_length=preconditioner_velocity_smooth_length,
            maximum_velocity_fields_update=maximum_velocity_fields_update,
            num_timesteps=num_timesteps,
            # Contrast map specifiers.
            contrast_order=contrast_order,
            spatially_varying_contrast_map=spatially_varying_contrast_map,
            contrast_iterations=contrast_iterations,
            sigma_contrast=sigma_contrast,
            contrast_smooth_length=contrast_smooth_length,
            # # vs. accuracy tradeoff.
            sigma_matching=sigma_matching,
            # Classification specifiers.
            classify_and_weight_voxels=classify_and_weight_voxels,
            sigma_artifact=sigma_artifact,
            sigma_background=sigma_background,
            artifact_prior=artifact_prior,
            background_prior=background_prior,
            # Initial values.
            initial_affine=initial_affine,
            initial_velocity_fields=initial_velocity_fields,
            initial_contrast_coefficients=initial_contrast_coefficients,
            # Diagnostic outputs.
            calibrate=calibrate,
            track_progress_every_n=track_progress_every_n,
        )

        # Fill unspecified parameters with presets if applicable.
        if preset is not None:
            preset_registration_parameters = get_registration_preset(preset)
            for registration_parameter in preset_registration_parameters.keys():
                # Override a registration_parameter with a preset value only if that parameter was not specified in this call (i.e. is None).
                if registration_parameters[registration_parameter] is None:
                    registration_parameters[registration_parameter] = preset_registration_parameters[registration_parameter]

        # Perform registration.
        lddmm_dict = lddmm_register(**registration_parameters)

        # Save registration parameters for the continue_registration method, with the initial_affine, initial_velocity_fields, and initial_contrast_coefficients updated.
        registration_parameters.update(
            initial_affine=lddmm_dict['affine'],
            initial_contrast_coefficients=lddmm_dict['contrast_coefficients'],
            initial_velocity_fields=lddmm_dict['velocity_fields'],
        )
        self._registration_parameters = registration_parameters

        # Update attributes.
        self._update_lddmm_attributes(lddmm_dict)

    
    def continue_registration(self, **registration_parameter_updates):
        """
        Continue registering with all the same registration parameters from the previous call to the register method, 
        but with initial_affine, initial_velocity_fields, and initial_contrast_coefficients set to the affine, velocity_fields, and contrast_coefficients 
        most recently calculated in the register method, updated by registration_parameter_updates.

        Kwargs:
            registration_parameter_updates (key-value pairs, optional): registration parameters provided as kwargs 
                to overwrite the most recent registration_parameters used in the register method. Defaults to {}.
        
        Raises:
            RuntimeError: Raised if self._registration_parameters are not set (as they should be after a call to the register method).
        """

        if self._registration_parameters is None:
            raise RuntimeError(f"The continue_registration method cannot be called from an object that has not performed the register method first.")

        # Update most recently saved registration parameters with user-provided kwargs.
        self._registration_parameters.update(registration_parameter_updates)

        # Continue the registration.
        self.register(**self._registration_parameters)


    def transform_image(self, subject, subject_resolution=1, output_resolution=None, output_shape=None, deform_to="template", extrapolation_fill_value=None, save_path=None):
        """
        Apply the transformation--computed by the last call to self.register--to subject, 
        deforming it into the space of deform_to.
        
        Args:
            subject (np.ndarray): The image to deform.
            subject_resolution (float, seq, optional): The resolution of subject in each dimension, or just one scalar to indicate isotropy. Defaults to 1.
            deform_to (str, optional): Either 'template' or 'target' indicating which to deform subject to match. Defaults to: "template".
            output_resolution (float, seq, optional): The resolution of the output deformed_subject in each dimension, 
                or just one scalar to indicate isotropy, or None to indicate the resolution of template or target based on deform_to. 
                Cannot be provided along with output_resolution. Defaults to None.
            output_shape (seq, optional): The shape of the output deformed_subject, or None to indicate the shape of the template or target based on deform_to. 
                Cannot be provided along with output_resolution. By default None.
            extrapolation_fill_value (float, NoneType, optional): The fill_value kwarg passed to scipy.interpolate.interpn; it should be background intensity. 
                If None, this is set to a low quantile of the subject's 10**-subject.ndim quantile to estimate background. Defaults to None.
            save_path (str, Path, optional): The full path to save the output to. Defaults to: None.
        
        Returns:
            np.ndarray: The result of deforming subject to match deform_to.
        """

        transformed_subject = lddmm_transform_image(
            subject=subject,
            subject_resolution=subject_resolution,
            output_resolution=output_resolution,
            output_shape=output_shape,
            deform_to=deform_to,
            extrapolation_fill_value=extrapolation_fill_value,
            **self.get_lddmm_dict(),
        )
        
        if save_path is not None:
            file_io.save(transformed_subject, save_path)

        return transformed_subject


    def transform_points(self, points, deform_to="template", save_path=None):
        """
        Apply the transformation--computed by the last call to self.register--to points, 
        deforming them into the space of deform_to.
        
        Args:
            points (np.ndarray): The points to deform. The last dimension should contain the coordinates for each point.
            deform_to (str, optional): Either 'template' or 'target' indicating which to deform points to match. Defaults to: "template".
            save_path (str, Path, optional): The full path to save the output to. Defaults to: None.
        
        Returns:
            np.ndarray: The result of deforming points to match deform_to, in the same shape as points was given.
        """

        transformed_points = lddmm_transform_points(
            points=points,
            deform_to=deform_to,
            affine_phi=self.affine_phi,
            phi_inv_affine_inv=self.phi_inv_affine_inv,
            template_resolution=self.template_resolution,
            target_resolution=self.target_resolution,
        )
        
        if save_path is not None:
            file_io.save(transformed_points, save_path)

        return transformed_points

    
    def save(self, file_path):
        """
        Save the entire instance of this Transform object (self) to file.
        
        Args:
            file_path (str, Path): The full path to save self to.
        """

        file_io.save_pickled(self, file_path)


    def load(self, file_path):
        """
        Load an entire instance of a Transform object from memory, as from a file created with the save method, 
        and transplants all of its writeable attributes into self.
        
        Args:
            file_path (str, Path): The full path that a Transform object was saved to.
        """

        transform = file_io.load_pickled(file_path)

        self.__dict__.update(transform.__dict__)


import sys
from setuptools import setup, find_packages
from pathlib import Path

# Variables defined in __version__.py.
version_info = {}
with open(Path(__file__).parent / 'ardent/__version__.py', 'r') as versionFile:
    exec(versionFile.read(), version_info)

# Descriptive text contained in README.md.
with open('README.md', 'r') as readmeFile:
    README = readmeFile.read()

# Requirements defined in requirements.txt.
with open(Path(__file__).parent / 'requirements.txt', 'r') as requirementsFile:
    install_requires = requirementsFile.readlines()

def check_python_version():
    """Raises SystemExit when the Python version is too low."""
    if sys.version_info < version_info['__min_python_version__']:
        raise SystemExit((f"Python >= {version_info['__min_python_version__'][0]}.{version_info['__min_python_version__'][1]} is required."
            f"\nYou are running Python {sys.version_info[0]}.{sys.version_info[1]}."))
check_python_version()

setup(
    name=version_info['__name__'],
    description=version_info['__description__'],
    long_description=README,
    long_description_content_type='text/markdown',
    version=version_info['__version__'],
    url=version_info['__url__'],
    author=version_info['__author__'],
    author_email=version_info['__author_email__'],
    python_requires=version_info['__python_requires__'],
    license=version_info['__license__'],
    classifiers=version_info['__classifiers__'],
    install_requires=install_requires,
    packages=find_packages(),
)